"""Module with all available Graph Neural Network models developed and used in the project"""

import inspect
import math
from typing import Any, Union

import lightning as L
import numpy as np
import torch
import torch.nn.functional as F
from lightning.pytorch.utilities.types import STEP_OUTPUT, OptimizerLRScheduler
from ogb.graphproppred.mol_encoder import AtomEncoder, BondEncoder
from torch.nn import SELU, BatchNorm1d, Dropout, Linear, ModuleList, ReLU, Sequential
from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts
from torch_geometric import nn as gnn
from torch_geometric.data.data import Data
from torch_geometric.nn import BatchNorm
from torchmetrics.functional import mean_absolute_percentage_error as mape

from .utils import rho_batch, vp_batch


class GNNePCSAFTL(L.LightningModule):
    """Graph neural network to predict ePCSAFT parameters with pytorch lightning."""

    def __init__(
        self,
        config: dict[str, Any],
    ):
        super().__init__()
        self.save_hyperparameters()
        self.config = config

        self.model = GNNePCSAFT(config)

    # pylint: disable=W0221
    def forward(
        self,
        x: torch.Tensor,
        edge_index: torch.Tensor,
        edge_attr: torch.Tensor,
        batch: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass of the model"""
        return self.model(x, edge_index, edge_attr, batch)

    def configure_optimizers(self) -> OptimizerLRScheduler:
        if self.config["optimizer"] == "adam":
            opt = torch.optim.AdamW(
                self.parameters(),
                lr=self.config["learning_rate"],
                weight_decay=self.config["weight_decay"],
                amsgrad=True,
                eps=1e-5,
            )
        elif self.config["optimizer"] == "sgd":
            opt = torch.optim.SGD(
                self.parameters(),
                lr=self.config["learning_rate"],
                momentum=0.0,
                weight_decay=0.0,
                nesterov=False,
            )
        else:
            raise ValueError(f"Unsupported optimizer: {self.config['optimizer']}.")
        return {
            "optimizer": opt,
            "lr_scheduler": {
                "scheduler": CosineAnnealingWarmRestarts(
                    opt, self.config["warmup_steps"], T_mult=2, eta_min=1e-6
                ),
                "interval": "epoch",
                "frequency": 10,
            },
        }

    def training_step(self, graphs, batch_idx) -> STEP_OUTPUT:  # pylint: disable=W0613
        if self.config["dataset"] in ("esper_assoc", "esper_assoc_only"):
            target: torch.Tensor = graphs.assoc
        else:
            target: torch.Tensor = graphs.para
        x, edge_index, edge_attr, batch = (
            graphs.x,
            graphs.edge_index,
            graphs.edge_attr,
            graphs.batch,
        )
        pred: torch.Tensor = self(x, edge_index, edge_attr, batch)
        ape = (pred - target) / target  # absolute percentage error
        zeros = torch.zeros_like(ape)
        loss = F.huber_loss(ape, zeros, delta=0.01)  # huber with ape
        loss_mape = mape(pred, target)
        self.log(
            "train_huber",
            loss,
            on_step=True,
            batch_size=target.shape[0],
            sync_dist=True,
        )
        self.log(
            "train_mape",
            loss_mape,
            on_step=True,
            batch_size=target.shape[0],
            sync_dist=True,
        )
        return loss

    def validation_step(  # pylint: disable=W0613,R0914
        self, graphs: Data, batch_idx: list, dataloader_idx: int = 0
    ) -> STEP_OUTPUT:
        metrics_dict = {}
        pred_para = self.model.pred_with_bounds(graphs).squeeze().detach()
        if self.config["num_para"] == 2:
            para_assoc = 10 ** (
                pred_para * torch.tensor([-1.0, 1.0], device=pred_para.device)
            )
            para_msigmae = graphs.para
        else:
            para_assoc = 10 ** (
                graphs.assoc * torch.tensor([-1.0, 1.0], device=pred_para.device)
            )
            para_msigmae = pred_para
        all_pred_para = (
            torch.hstack([para_msigmae, para_assoc, graphs.munanb])
            .cpu()
            .to(torch.float64)
            .tolist()
        )
        pred_rho = rho_batch(all_pred_para, graphs.rho)
        pred_vp = vp_batch(all_pred_para, graphs.vp)
        rho = [rho[:, -1] for rho in graphs.rho if rho.shape[0] > 0]
        vp = [vp[:, -1] for vp in graphs.vp if vp.shape[0] > 0]
        mape_den = []
        for pred, exp in zip(pred_rho, rho):
            mape_den += [np.mean(np.abs(pred - exp) / exp).item()]
        mape_den = np.asarray(mape_den).mean().item()
        mape_vp = []
        for pred, exp in zip(pred_vp, vp):
            mape_vp += [np.mean(np.abs(pred - exp) / exp).item()]
        mape_vp = np.asarray(mape_vp).mean().item()
        metrics_dict.update(
            {
                "mape_den": mape_den,
                "mape_vp": mape_vp,
            }
        )

        self.log_dict(
            metrics_dict, on_step=False, on_epoch=True, batch_size=1, sync_dist=True
        )
        return metrics_dict

    def test_step(self, graphs, batch_idx, dataloader_idx=0) -> STEP_OUTPUT:
        return self.validation_step(graphs, batch_idx, dataloader_idx)


class GNNePCSAFT(torch.nn.Module):  # pylint: disable=R0902
    """Graph neural network to predict ePCSAFT parameters"""

    def __init__(self, config: dict):
        super().__init__()

        self.convs = ModuleList()
        self.batch_norms = ModuleList()
        self.lower_bounds = torch.tensor(
            [1.0, 1.9, 50.0, -1 * math.log10(0.9), math.log10(200.0)]
        )
        self.upper_bounds = torch.tensor(
            [25.0, 4.5, 550.0, -1 * math.log10(0.0001), math.log10(5000.0)]
        )
        self.num_para = config["num_para"]

        self.node_embed = AtomEncoder(config["hidden_dim"])
        self.edge_embed = BondEncoder(config["hidden_dim"])
        self.dropout = Dropout(p=config["dropout"])
        self.global_pool = get_global_pool(config)
        self.global_pool_type = config["global_pool"]

        for _ in range(config["propagation_depth"]):
            conv = get_conv(config)
            self.convs.append(conv)
            self.batch_norms.append(BatchNorm(config["hidden_dim"]))

        self.mlp = Sequential(
            Linear(config["hidden_dim"], config["hidden_dim"] // 2),
            BatchNorm1d(config["hidden_dim"] // 2),
            ReLU(),
            Linear(config["hidden_dim"] // 2, config["hidden_dim"] // 4),
            BatchNorm1d(config["hidden_dim"] // 4),
            ReLU(),
            Linear(config["hidden_dim"] // 4, config["num_para"]),
        )

    def forward(
        self,
        x: torch.Tensor,
        edge_index: torch.Tensor,
        edge_attr: torch.Tensor,
        batch: Union[torch.Tensor, None],
    ) -> torch.Tensor:
        """Forward pass of the model"""

        x = self.node_embed(x)
        edge_attr = self.edge_embed(edge_attr)

        for conv, batch_norm in zip(self.convs, self.batch_norms):
            x = self.dropout(x)
            # check if edge_attr is a keyword argument in conv
            if "edge_attr" in inspect.signature(conv.forward).parameters:
                x = F.relu(
                    batch_norm(conv(x=x, edge_index=edge_index, edge_attr=edge_attr))
                )
            else:
                x = F.relu(batch_norm(conv(x=x, edge_index=edge_index)))

        if batch is not None:
            x = self.global_pool(x, batch)
        elif self.global_pool_type == "mean":
            x = x.mean(dim=0, keepdim=True)
        elif self.global_pool_type == "max":
            x = x.max(dim=0, keepdim=True).values
        elif self.global_pool_type == "add":
            x = x.sum(dim=0, keepdim=True)
        x = self.mlp(x)
        return x

    def pred_with_bounds(self, data: Data):
        """Forward pass of the model with bounds."""

        x, edge_index, edge_attr, batch = (
            data.x,
            data.edge_index,
            data.edge_attr,
            data.batch,
        )

        if (
            isinstance(x, torch.Tensor)
            and isinstance(edge_index, torch.Tensor)
            and isinstance(edge_attr, torch.Tensor)
        ):

            params = self.forward(x, edge_index, edge_attr, batch)
            upper_bounds = (
                self.upper_bounds[:3] if self.num_para == 3 else self.upper_bounds[3:]
            ).to(device=x.device)
            lower_bounds = (
                self.lower_bounds[:3] if self.num_para == 3 else self.lower_bounds[3:]
            ).to(device=x.device)

            return params.clip(lower_bounds, upper_bounds)
        raise ValueError("Invalid input data")


class HabitchNN(torch.nn.Module):
    """
    Neural network to predict PCSAFT parameters
    from (Habicht; Brandenbusch; Sadowski, 2023, 10.1016/j.fluid.2022.113657).
    """

    def __init__(self, input_dim: int):
        super().__init__()
        self.lower_bounds = torch.tensor([1.0, 1.9, 50.0])
        self.upper_bounds = torch.tensor([25.0, 4.5, 550.0])

        self.mlp = Sequential(
            Linear(input_dim, 2048),
            SELU(),
            Dropout(p=0.1),
            Linear(2048, 1024),
            SELU(),
            Dropout(p=0.1),
            Linear(1024, 1024),
            SELU(),
            Dropout(p=0.1),
            Linear(1024, 512),
            SELU(),
            Dropout(p=0.1),
            Linear(512, 128),
            SELU(),
            Dropout(p=0.1),
            Linear(128, 32),
            SELU(),
            Dropout(p=0.1),
            Linear(32, 3),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the model"""
        return self.mlp(x)

    def pred_with_bounds(self, graphs: Data):
        """Forward pass of the model with bounds."""

        ecfp, mw, atom_count, ring_count, rbonds_count = (
            graphs.ecfp,
            graphs.mw,
            graphs.atom_count,
            graphs.ring_count,
            graphs.rbond_count,
        )
        x = torch.hstack((ecfp, mw, atom_count, ring_count, rbonds_count))

        params = self.forward(x)
        upper_bounds = self.upper_bounds.to(device=x.device)
        lower_bounds = self.lower_bounds.to(device=x.device)

        return params.clip(lower_bounds, upper_bounds)


class HabitchNNL(L.LightningModule):
    """
    Neural network to predict PCSAFT parameters
    from (Habicht; Brandenbusch; Sadowski, 2023, 10.1016/j.fluid.2022.113657).
    """

    def __init__(self, config: dict):
        super().__init__()
        self.save_hyperparameters()

        self.config = config

        self.model = HabitchNN(3085)

    # pylint: disable=W0221
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the model"""
        return self.model(x)

    def configure_optimizers(self) -> OptimizerLRScheduler:
        if self.config["optimizer"] == "adam":
            opt = torch.optim.AdamW(
                self.parameters(),
                lr=self.config["learning_rate"],
                weight_decay=self.config["weight_decay"],
                amsgrad=True,
                eps=1e-5,
            )
        elif self.config["optimizer"] == "sgd":
            opt = torch.optim.SGD(
                self.parameters(),
                lr=self.config["learning_rate"],
                momentum=0.0,
                weight_decay=0.0,
                nesterov=False,
            )
        else:
            raise ValueError(f"Unsupported optimizer: {self.config['optimizer']}.")
        return {
            "optimizer": opt,
            "lr_scheduler": {
                "scheduler": CosineAnnealingWarmRestarts(
                    opt, self.config["warmup_steps"]
                ),
                "interval": "step",
                "frequency": 1,
            },
        }

    def training_step(self, graphs, batch_idx) -> STEP_OUTPUT:  # pylint: disable=W0613
        """Training step of the model"""
        target: torch.Tensor = graphs.para

        ecfp, mw, atom_count, ring_count, rbonds_count = (
            graphs.ecfp,
            graphs.mw,
            graphs.atom_count,
            graphs.ring_count,
            graphs.rbond_count,
        )
        x = torch.hstack((ecfp, mw, atom_count, ring_count, rbonds_count))

        pred: torch.Tensor = self(x)
        ape = (pred - target) / target  # absolute percentage error
        zeros = torch.zeros_like(ape)
        loss = F.huber_loss(ape, zeros, delta=0.01)  # huber with ape
        loss_mape = mape(pred, target)
        self.log(
            "train_huber",
            loss,
            on_step=True,
            batch_size=target.shape[0],
            sync_dist=True,
        )
        self.log(
            "train_mape",
            loss_mape,
            on_step=True,
            batch_size=target.shape[0],
            sync_dist=True,
        )
        return loss

    def validation_step(  # pylint: disable=W0613,R0914
        self, graphs, batch_idx, dataloader_idx=0
    ) -> STEP_OUTPUT:
        metrics_dict = {}
        para_msigmae: torch.Tensor = (
            self.model.pred_with_bounds(graphs).squeeze().detach()
        )
        para_assoc = 10 ** (
            graphs.assoc * torch.tensor([-1.0, 1.0], device=para_msigmae.device)
        )
        pred_para = (
            torch.hstack([para_msigmae, para_assoc, graphs.munanb])
            .cpu()
            .to(torch.float64)
            .tolist()
        )
        pred_rho = rho_batch(pred_para, graphs.rho)
        pred_vp = vp_batch(pred_para, graphs.vp)
        rho = [rho[:, -1] for rho in graphs.rho if rho.shape[0] > 0]
        vp = [vp[:, -1] for vp in graphs.vp if vp.shape[0] > 0]
        mape_den = []
        for pred, exp in zip(pred_rho, rho):
            mape_den += [np.mean(np.abs(pred - exp) / exp).item()]
        mape_den = np.asarray(mape_den).mean().item()
        mape_vp = []
        for pred, exp in zip(pred_vp, vp):
            mape_vp += [np.mean(np.abs(pred - exp) / exp).item()]
        mape_vp = np.asarray(mape_vp).mean().item()
        metrics_dict.update(
            {
                "mape_den": mape_den,
                "mape_vp": mape_vp,
            }
        )

        self.log_dict(
            metrics_dict, on_step=False, on_epoch=True, batch_size=1, sync_dist=True
        )
        return metrics_dict

    def test_step(self, graphs, batch_idx, dataloader_idx=0) -> STEP_OUTPUT:
        return self.validation_step(graphs, batch_idx, dataloader_idx)


def get_conv(config: dict):  # pylint: disable=R0911,R0912
    """Returns the convolution layer."""
    aggregators = ["mean", "min", "max", "std"]
    scalers = ["identity", "amplification", "attenuation"]
    if config["conv"] == "PNA":  # 2020, https://doi.org/10.48550/arXiv.2004.05718
        return gnn.PNAConv(
            in_channels=config["hidden_dim"],
            out_channels=config["hidden_dim"],
            aggregators=aggregators,
            scalers=scalers,
            deg=torch.tensor(config["deg"], dtype=torch.long),
            edge_dim=config["hidden_dim"],
            towers=config["towers"],
            pre_layers=config["pre_layers"],
            post_layers=config["post_layers"],
            divide_input=True,
        )

    if config["conv"] == "GCN":  # 2016-2017, https://doi.org/10.48550/arXiv.1609.02907
        return gnn.GCNConv(
            in_channels=config["hidden_dim"],
            out_channels=config["hidden_dim"],
            add_self_loops=config["add_self_loops"],
        )

    if config["conv"] == "GAT":  # 2017-2018, https://doi.org/10.48550/arXiv.1710.10903
        assert (
            config["hidden_dim"] % config["heads"] == 0
        ), "hidden_dim must be divisible by heads"

        return gnn.GATConv(
            in_channels=config["hidden_dim"],
            out_channels=config["hidden_dim"] // config["heads"],
            heads=config["heads"],
            concat=True,
            dropout=config["dropout"],
            edge_dim=config["hidden_dim"],
            add_self_loops=config["add_self_loops"],
        )

    if (
        config["conv"] == "GATv2"
    ):  # 2021-2022, https://doi.org/10.48550/arXiv.2105.14491
        assert (
            config["hidden_dim"] % config["heads"] == 0
        ), "hidden_dim must be divisible by heads"
        return gnn.GATv2Conv(
            in_channels=config["hidden_dim"],
            out_channels=config["hidden_dim"] // config["heads"],
            heads=config["heads"],
            concat=True,
            dropout=config["dropout"],
            edge_dim=config["hidden_dim"],
            add_self_loops=config["add_self_loops"],
        )

    if (
        config["conv"] == "Transformer"
    ):  # 2020-2021, https://doi.org/10.48550/arXiv.2009.03509
        assert (
            config["hidden_dim"] % config["heads"] == 0
        ), "hidden_dim must be divisible by heads"
        return gnn.TransformerConv(
            in_channels=config["hidden_dim"],
            out_channels=config["hidden_dim"] // config["heads"],
            heads=config["heads"],
            concat=True,
            dropout=config["dropout"],
            edge_dim=config["hidden_dim"],
        )

    if config["conv"] == "SAGE":  # 2017-2018, https://doi.org/10.48550/arXiv.1706.02216
        return gnn.SAGEConv(
            in_channels=config["hidden_dim"],
            out_channels=config["hidden_dim"],
            aggr=aggregators,
        )

    if config["conv"] == "GIN":  # 2018-2019, https://doi.org/10.48550/arXiv.1810.00826
        return gnn.GINConv(
            nn=Sequential(
                Linear(config["hidden_dim"], config["hidden_dim"]),
                ReLU(),
                Linear(config["hidden_dim"], config["hidden_dim"]),
            ),
            train_eps=False,
        )

    if config["conv"] == "GINE":  # 2019-2020, https://doi.org/10.48550/arXiv.1905.12265
        return gnn.GINEConv(
            nn=Sequential(
                Linear(config["hidden_dim"], config["hidden_dim"]),
                ReLU(),
                Linear(config["hidden_dim"], config["hidden_dim"]),
            ),
            train_eps=False,
            edge_dim=config["hidden_dim"],
        )

    if config["conv"] == "Edge":  # 2018-2019, https://doi.org/10.48550/arXiv.1801.07829
        return gnn.EdgeConv(
            nn=Sequential(
                Linear(2 * config["hidden_dim"], config["hidden_dim"]),
                ReLU(),
                Linear(config["hidden_dim"], config["hidden_dim"]),
            ),
            aggr="max",
        )

    if (
        config["conv"] == "GatedGraph"
    ):  # 2015-2017, https://doi.org/10.48550/arXiv.1511.05493
        return gnn.GatedGraphConv(
            out_channels=config["hidden_dim"],
            num_layers=config["num_layers"],
        )

    if (
        config["conv"] == "Graph"
    ):  # 2018-2021, https://doi.org/10.48550/arXiv.1810.02244
        return gnn.GraphConv(
            in_channels=config["hidden_dim"],
            out_channels=config["hidden_dim"],
        )

    if (
        config["conv"] == "ARMA"
    ):  # 2019-2021, https://doi.org/10.1109/TPAMI.2021.3054830
        return gnn.ARMAConv(
            in_channels=config["hidden_dim"],
            out_channels=config["hidden_dim"],
            num_stacks=config["num_stacks"],
            num_layers=config["num_layers"],
            dropout=config["dropout"],
        )

    if config["conv"] == "SG":  # 2019, https://doi.org/10.48550/arXiv.1902.07153
        return gnn.SGConv(
            in_channels=config["hidden_dim"],
            out_channels=config["hidden_dim"],
            add_self_loops=config["add_self_loops"],
        )

    raise ValueError(f"Unsupported convolution: {config['conv']}.")


def get_global_pool(config: dict):
    """Returns the global pooling layer."""
    if config["global_pool"] == "mean":
        return gnn.aggr.MeanAggregation()
    if config["global_pool"] == "max":
        return gnn.aggr.MaxAggregation()
    if config["global_pool"] == "add":
        return gnn.aggr.SumAggregation()
    raise ValueError(f"Unsupported global pooling: {config['global_pool']}.")


def create_model(config: dict[str, Any], deg: list[int]):
    """Creates a model, as specified by the config."""
    config["deg"] = deg

    if config["model"].lower() == "gnn":
        return GNNePCSAFTL(config)
    if config["model"].lower() == "habitch":
        return HabitchNNL(config)
    raise ValueError(f"Unsupported model: {config['model']}.")



"""Module to be used for ThermoML Archive dataset building"""

import os.path as osp
from urllib.parse import quote
from urllib.request import HTTPError, urlopen

import polars as pl
from absl import app, flags, logging


def ramirez(root: str, save_dir: str):
    """
    Preprocess `Ramírez-Vélez et al. (2022, doi: 10.1002/aic.17722)` data
    and saves into dataframe file used in data loading.

    PARAMETERS
    ----------
    root (str) – Path where the ramirez data file is.

    save_dir (str) - Directory to save output file.
    """
    data = pl.read_csv(root, has_header=True, separator=";")

    inchis = data["name"].map_elements(to_inchi).rename("inchi")
    data = data.with_columns(inchis)
    file_path = osp.join(save_dir, "ramirez2022/raw/data.parquet")
    data.write_parquet(file_path)


def to_inchi(ids: str) -> str:
    "Tries to return InChI from input compound name with pubchem or cactus api."
    try:
        url = (
            "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/name/"
            + quote(ids)
            + "/property/InChI/TXT"
        )
        with urlopen(url) as ans:
            ans = ans.read().decode("utf8").rstrip()
            ans = ans.split("\n")[0]

    except (TypeError, ValueError, HTTPError):
        try:
            url = (
                "http://cactus.nci.nih.gov/chemical/structure/" + quote(ids) + "/inchi"
            )
            with urlopen(url) as ans:
                ans = ans.read().decode("utf8").rstrip()
                ans = ans.split("\n")[0]
        except (TypeError, ValueError, HTTPError):
            print("not ok:", ids)
            ans = ""
    return ans


FLAGS = flags.FLAGS

flags.DEFINE_string("root", osp.dirname(__file__), "Path to data directory.")
flags.DEFINE_string("save_dir", osp.dirname(__file__), "Path to output directory")


def main(argv):
    """Execution from command line"""
    if len(argv) > 1:
        raise app.UsageError("Too many command-line arguments.")

    logging.info("Calling preprocess!")
    ramirez(osp.join(FLAGS.root, "ramirez2022/raw/RAMIREZ2022.csv"), FLAGS.save_dir)
    logging.info("Ramírez-Vélez dataset preprocessed!")


if __name__ == "__main__":
    app.run(main)



"""Module for molecular graph dataset building."""

import polars as pl
import torch
from torch_geometric.data import InMemoryDataset

from .graph import from_InChI
from .rdkit_util import mw


class ThermoMLDataset(InMemoryDataset):
    """
    Molecular Graph dataset creator/manipulator with `ThermoML Archive` experimental data.

    PARAMETERS
    ----------
    root (str, optional) – Root directory where the dataset should be saved. (optional: None).

    transform (callable, optional) – A function/transform that takes in an Data object and
    returns a transformed version. The data object will be transformed
    before every access. (default: None).

    pre_transform (callable, optional) – A function/transform that takes in
    an Data object and returns a transformed version. The data object will be
    transformed before being saved to disk. (default: None).

    pre_filter (callable, optional) – A function that takes in an Data object
    and returns a boolean value, indicating whether the data object should be
    included in the final dataset. (default: None).

    log (bool, optional) – Whether to print any console output while downloading
    and processing the dataset. (default: True).

    """

    def __init__(
        self,
        root,
        transform=None,
        pre_transform=None,
        pre_filter=None,
    ):
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[0], weights_only=False)

    @property
    def raw_file_names(self):
        return ["pure.parquet"]

    @property
    def processed_file_names(self):
        return ["tml_graph_data.pt"]

    def download(self):
        raise ValueError(
            f"No url to download from. Provide data at {self.raw_paths[0]}."
        )

    def process(self):
        datalist = []

        with open(self.raw_paths[0], "rb") as f:
            data = pl.read_parquet(f)
        inchis = data.unique("inchi1")["inchi1"].to_list()

        for inchi in inchis:
            try:
                graph = from_InChI(inchi)
            except (TypeError, ValueError) as e:
                print(f"Error for InChI:\n {inchi}", e, sep="\n\n", end="\n\n")
                continue

            graph.vp = (
                data.filter(pl.col("inchi1") == inchi, pl.col("tp") == 3)
                .select("TK", "PPa", "phase", "tp", "m")
                .to_numpy()
            )
            rho = (
                data.filter(pl.col("inchi1") == inchi, pl.col("tp") == 1)
                .select("TK", "PPa", "phase", "tp", "m")
                .to_numpy()
            )

            rho[:, -1] *= 1000 / mw(inchi)  # convert to mol/ m³
            graph.rho = rho

            datalist.append(graph)

        torch.save(self.collate(datalist), self.processed_paths[0])


class Ramirez(InMemoryDataset):
    """
    Molecular Graph dataset creator/manipulator with `ePC-SAFT` parameters from
    `Ramírez-Vélez et al. (2022, doi: 10.1002/aic.17722)`.

    PARAMETERS
    ----------
    root (str, optional) – Root directory where the dataset should be saved. (optional: None).

    transform (callable, optional) – A function/transform that takes in an Data object and
    returns a transformed version. The data object will be transformed
    before every access. (default: None).

    pre_transform (callable, optional) – A function/transform that takes in
    an Data object and returns a transformed version. The data object will be
    transformed before being saved to disk. (default: None).

    pre_filter (callable, optional) – A function that takes in an Data object
    and returns a boolean value, indicating whether the data object should be
    included in the final dataset. (default: None).

    log (bool, optional) – Whether to print any console output while downloading
    and processing the dataset. (default: True).

    """

    def __init__(self, root, transform=None, pre_transform=None, pre_filter=None):
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[0], weights_only=False)

    @property
    def raw_file_names(self):
        return ["data.parquet"]

    @property
    def processed_file_names(self):
        return ["ra_graph_data.pt"]

    def download(self):
        raise ValueError(
            f"No url to download from. Provide data at {self.raw_paths[0]}."
        )

    def process(self):
        datalist = []
        data = pl.read_parquet(self.raw_paths[0])

        for row in data.iter_rows():
            inchi = row[-1]
            para = row[3:6]
            try:
                graph = from_InChI(inchi)
            except (TypeError, ValueError) as e:
                print(f"Error for InChI:\n {inchi}", e, sep="\n\n", end="\n\n")
                continue

            graph.para = torch.tensor(para)
            graph.critic = torch.tensor(row[1:3])
            datalist.append(graph)

        torch.save(self.collate(datalist), self.processed_paths[0])


class Esper(InMemoryDataset):
    """
    Molecular Graph dataset creator/manipulator with `ePC-SAFT` parameters from
    `Esper et al. (2023, doi: 10.1021/acs.iecr.3c02255)`.

    PARAMETERS
    ----------
    root (str, optional) – Root directory where the dataset should be saved. (optional: None).

    transform (callable, optional) – A function/transform that takes in an Data object and
    returns a transformed version. The data object will be transformed
    before every access. (default: None).

    pre_transform (callable, optional) – A function/transform that takes in
    an Data object and returns a transformed version. The data object will be
    transformed before being saved to disk. (default: None).

    pre_filter (callable, optional) – A function that takes in an Data object
    and returns a boolean value, indicating whether the data object should be
    included in the final dataset. (default: None).

    log (bool, optional) – Whether to print any console output while downloading
    and processing the dataset. (default: True).

    """

    def __init__(self, root, transform=None, pre_transform=None, pre_filter=None):
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[0], weights_only=False)

    @property
    def raw_file_names(self):
        return ["SI_pcp-saft_parameters.csv"]

    @property
    def processed_file_names(self):
        return ["es_graph_data.pt"]

    def download(self):
        raise ValueError(
            f"No url to download from. Provide data at {self.raw_paths[0]}."
        )

    def process(self):
        datalist = []
        dtype = torch.float32
        data = pl.read_csv(self.raw_paths[0], separator="	")

        for row in data.iter_rows():

            inchi = row[2]
            para = [value if value else 0.0001 for value in row[8:11]]
            assoc = row[12:14] if all(row[12:14]) else [0.0001, 200]
            munanb = [value if value else 0.0 for value in row[11:12] + row[14:16]]
            try:
                graph = from_InChI(inchi)
            except (TypeError, ValueError) as e:
                print(f"Error for InChI:\n {inchi}", e, sep="\n\n", end="\n\n")
                continue

            graph.para = torch.tensor([para], dtype=dtype)
            graph.assoc = torch.tensor([assoc], dtype=dtype).log10().abs()
            graph.munanb = torch.tensor([munanb], dtype=dtype)
            datalist.append(graph)

        torch.save(self.collate(datalist), self.processed_paths[0])



"""Module to get configs for model hyperparameters in parallel training"""

from typing import Any


def get_configs() -> list[dict[str, Any]]:
    """Get the hyperparameter configurations."""
    configs = [
        {
            "model_name": "pna_msigmae_1.0",
            "model": "gnn",
            "conv": "PNA",
            "dropout": 0.0,
            "propagation_depth": 6,
            "hidden_dim": 256,
            "post_layers": 4,
            "pre_layers": 2,
            "towers": 1,
            "num_para": 3,
            "dataset": "esper",
            "batch_size": 512,
            "checkpoint": "wildson/gnn-pc-saft/model-gq9gcgmb:v3",
            "resume_id": "",
            "num_train_steps": 425_000,
        },
        {
            "model_name": "gatv2_msigmae_1.0",
            "model": "gnn",
            "conv": "GATv2",
            "dropout": 0.0,
            "propagation_depth": 3,
            "hidden_dim": 512,
            "heads": 8,
            "num_para": 3,
            "dataset": "esper",
            "batch_size": 512,
            "checkpoint": "wildson/gnn-pc-saft/model-54730h51:v0",
            "resume_id": "",
            "num_train_steps": 600_000,
        },
        {
            "model_name": "gatv2_assoc_1.0",
            "model": "gnn",
            "conv": "GATv2",
            "dropout": 0.0,
            "propagation_depth": 3,
            "hidden_dim": 512,
            "heads": 8,
            "num_para": 2,
            "dataset": "esper_assoc_only",
            "batch_size": 387 // 4 + 1,
            "checkpoint": "wildson/gnn-pc-saft/model-zme6255l:v1",
            "resume_id": "",
            "num_train_steps": 575_000,
        },
    ]
    return configs



"""Model with important functions to help model training"""

import multiprocessing as mp
import os.path as osp
import time
from tempfile import TemporaryDirectory
from typing import Any, List, Union

import numpy as np
import torch
import torch_geometric.transforms as T
import xgboost as xgb
from absl import logging
from lightning import LightningModule, Trainer
from lightning.pytorch.callbacks import Callback
from ray import tune
from ray.tune.experiment.trial import Trial
from sklearn.ensemble import RandomForestRegressor
from torch_geometric.data import Data
from torch_geometric.transforms import BaseTransform
from torch_geometric.utils import degree
from xgboost import Booster

from ..data.graph import assoc_number
from ..data.graphdataset import Esper, Ramirez, ThermoMLDataset
from ..epcsaft.utils import pure_den_feos, pure_vp_feos

# PCSAFT parameters bounds
params_lower_bound = np.array([1.0, 1.9, 50.0, 1e-4, 200.0, 0, 0, 0])
params_upper_bound = np.array([25.0, 4.5, 550.0, 0.9, 5000.0, np.inf, np.inf, np.inf])


def calc_deg(dataset: str, workdir: str) -> List:
    """Calculates deg for `PNA` conv."""
    if dataset == "ramirez":
        path = osp.join(workdir, "data/ramirez2022")
        train_dataset = Ramirez(path)
    elif dataset in ("esper", "esper_assoc", "esper_assoc_only"):
        path = osp.join(workdir, "data/esper2023")
        train_dataset = Esper(path)
    else:
        raise ValueError(
            f"dataset is either ramirez or esper, got >>> {dataset} <<< instead"
        )
    # Compute the maximum in-degree in the training data.
    max_degree = -1
    for data in train_dataset:
        d = degree(data.edge_index[1], num_nodes=data.num_nodes, dtype=torch.long)
        max_degree = max(max_degree, int(d.max()))

    # Compute the in-degree histogram tensor
    deg = torch.zeros(max_degree + 1, dtype=torch.long)
    for data in train_dataset:
        d = degree(data.edge_index[1], num_nodes=data.num_nodes, dtype=torch.long)
        deg += torch.bincount(d, minlength=deg.numel())
    return deg.tolist()


def rhovp_data(parameters: List, rho: np.ndarray, vp: np.ndarray):
    """Calculates density and vapor pressure with ePC-SAFT"""

    den_array = rho_single((parameters, rho))
    vp_array = vp_single((parameters, vp))

    return den_array, vp_array


# pylint: disable=R0903
class TransformParameters(BaseTransform):
    "To add parameters to test dataset."

    def __init__(self, para_data: dict) -> None:
        super().__init__()
        self.para_data = para_data

    def forward(self, data: Any) -> Any:
        if data.InChI in self.para_data:
            data.para, data.assoc, data.munanb = self.para_data[data.InChI]
        else:
            data.para, data.assoc = (
                torch.zeros(1, 3),
                torch.zeros(1, 2),
            )
            data.munanb = torch.tensor(
                [(0,) + assoc_number(data.InChI)], dtype=torch.float32
            )
        return data


def build_test_dataset(
    workdir: str,
    train_dataset: Union[Esper, Ramirez],
    transform: Union[None, BaseTransform] = None,
) -> tuple[ThermoMLDataset, ThermoMLDataset]:
    "Builds test dataset."

    para_data = {}
    if isinstance(train_dataset, (Esper,)):
        for graph in train_dataset:
            para_data[graph.InChI] = (
                graph.para,
                graph.assoc,
                graph.munanb,
            )
    if transform:
        transform = T.Compose([TransformParameters(para_data), transform])
    else:
        transform = TransformParameters(para_data)
    tml_dataset = ThermoMLDataset(
        osp.join(workdir, "data/thermoml"), transform=transform
    )

    val_msigmae_idx: List[int] = []
    train_idx: List[int] = []
    # separate test and val dataset
    for idx, graph in enumerate(tml_dataset):
        if graph.InChI not in para_data and graph.munanb[0, -1] == 0:
            val_msigmae_idx.append(idx)
        if graph.InChI in para_data:
            train_idx.append(idx)
    val_msigmae_dataset = tml_dataset[val_msigmae_idx]
    train_val_dataset = tml_dataset[train_idx]
    return val_msigmae_dataset, train_val_dataset  # type: ignore


def build_train_dataset(workdir, dataset, transform=None) -> Union[Esper, Ramirez]:
    "Builds train dataset."
    if dataset == "ramirez":
        path = osp.join(workdir, "data/ramirez2022")
        return Ramirez(path, transform=transform)
    if dataset == "esper":
        path = osp.join(workdir, "data/esper2023")
        return Esper(path, transform=transform)
    if dataset == "esper_assoc":
        path = osp.join(workdir, "data/esper2023")
        train_dataset = Esper(path, transform=transform)
        assoc_idx = []
        non_assoc_idx = []
        for i, graph in enumerate(train_dataset):
            if all(graph.munanb[0, 1:] > 0):
                assoc_idx.append(i)
            if all(graph.munanb[0, 1:] == 0):
                non_assoc_idx.append(i)
        dataset_idxs = assoc_idx * 4 + non_assoc_idx
        return train_dataset[dataset_idxs]  # type: ignore
    if dataset == "esper_assoc_only":
        path = osp.join(workdir, "data/esper2023")
        train_dataset = Esper(path, transform=transform)
        as_idx = []
        for i, graph in enumerate(train_dataset):
            if all(graph.munanb[0, 1:] > 0):
                as_idx.append(i)
        return train_dataset[as_idx]  # type: ignore
    raise ValueError(
        f"dataset is either ramirez, esper, esper_assoc \
              or esper_assoc_only, got >>> {dataset} <<< instead"
    )


class EpochTimer(Callback):
    "Elapsed time counter."

    start_time: float = time.time()

    def on_train_epoch_start(
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        self.start_time = time.time()

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        end_time = time.time()

        elapsed_time = end_time - self.start_time
        logging.log_first_n(
            logging.INFO, "Elapsed time %.4f min.", 20, elapsed_time / 60
        )


class CustomRayTrainReportCallback(Callback):
    "Lightning Callback for Ray Tune reporting."

    def on_validation_end(self, trainer, pl_module):

        with TemporaryDirectory() as tmpdir:
            # Fetch metrics
            metrics = trainer.callback_metrics
            metrics = {k: v.item() for k, v in metrics.items()}

            # Add customized metrics
            metrics["epoch"] = trainer.current_epoch
            metrics["step"] = trainer.global_step

            trial_id = tune.get_context().get_trial_id()
            # Save model checkpoint file to tmpdir
            ckpt_path = osp.join(tmpdir, f"{trial_id}.ckpt")
            trainer.save_checkpoint(ckpt_path, weights_only=False)
            checkpoint = tune.Checkpoint.from_directory(tmpdir)

            # Report to train session
            tune.report(metrics=metrics, checkpoint=checkpoint)


class TrialTerminationReporter(tune.JupyterNotebookReporter):
    """Reporter for Ray to report only when trial is terminated"""

    def __init__(self):
        super().__init__()
        self.num_terminated = 0

    def should_report(self, trials, done=False):
        """
        Reports only on trial termination events.
        It does so by tracking increase in number of trials terminated.
        """
        old_num_terminated = self.num_terminated
        self.num_terminated = len([t for t in trials if t.status == Trial.TERMINATED])
        return self.num_terminated > old_num_terminated


class CustomStopper(tune.Stopper):
    """Custom ray tune experiment/trial stopper"""

    def __init__(self, max_iter: int):
        self.should_stop = False
        self.max_iter = max_iter

    def __call__(self, trial_id, result):
        return result["training_iteration"] >= self.max_iter

    def stop_all(self):
        return False


def rho_single(args: tuple[List, np.ndarray]) -> np.ndarray:
    """Calculates density with ePC-SAFT for a single pararameter"""
    parameters, states = args
    rho_for_all_states = []

    for state in states:
        try:
            rho_for_state = pure_den_feos(parameters, state)
        except (AssertionError, RuntimeError):
            rho_for_state = 0.0
        rho_for_all_states += [rho_for_state]
    return np.asarray(rho_for_all_states)


def rho_batch(
    parameters_batch: List[List[Any]], states_batch: List[np.ndarray]
) -> List[np.ndarray]:
    """
    Calculates density with ePC-SAFT
    for a batch of parameters
    using nested multiprocessing
    """
    args_list = [
        (para, states)
        for para, states in zip(parameters_batch, states_batch)
        if states.shape[0] > 0
    ]
    ctx = mp.get_context("spawn")
    with ctx.Pool(processes=ctx.cpu_count() // 2) as pool:
        den = pool.map(rho_single, args_list)
    return den


def vp_single(args: tuple[List, np.ndarray]) -> np.ndarray:
    """Calculates vapor pressure with ePC-SAFT for a single parameter"""
    parameters, states = args
    vp_for_all_states = []
    for state in states:
        try:
            vp_for_state = pure_vp_feos(parameters, state)
        except (AssertionError, RuntimeError):
            vp_for_state = 0.0
        vp_for_all_states += [vp_for_state]
    return np.asarray(vp_for_all_states)


def vp_batch(
    parameters_batch: List[List[Any]], states_batch: List[np.ndarray]
) -> List[np.ndarray]:
    """
    Calculates vapor pressure with ePC-SAFT
    for a batch of parameters
    using nested multiprocessing
    """
    args_list = [
        (para, states)
        for para, states in zip(parameters_batch, states_batch)
        if states.shape[0] > 0
    ]
    ctx = mp.get_context("spawn")
    with ctx.Pool(processes=ctx.cpu_count() // 2) as pool:
        vp = pool.map(vp_single, args_list)
    return vp


def rf_xgb_evaluation(
    graphs: Data, model: Union[RandomForestRegressor, Booster]
) -> tuple[float, float]:
    """Evaluation function for RF and XGBoost models"""
    if isinstance(model, Booster):
        x = xgb.DMatrix(
            torch.hstack(
                (
                    graphs.ecfp,
                    graphs.mw,
                    graphs.atom_count,
                    graphs.ring_count,
                    graphs.rbond_count,
                )
            ).numpy()
        )
        pred_msigmae = model.predict(x)
    else:
        x = torch.hstack(
            (
                graphs.ecfp,
                graphs.mw,
                graphs.atom_count,
                graphs.ring_count,
                graphs.rbond_count,
            )
        ).numpy()
        pred_msigmae = model.predict(x)

    assert isinstance(pred_msigmae, np.ndarray)
    assert pred_msigmae.shape == graphs.para.numpy().shape
    para_assoc = 10 ** (graphs.assoc.numpy() * np.array([-1.0, 1.0]))
    pred_params = np.hstack([pred_msigmae, para_assoc, graphs.munanb.numpy()])
    pred_params.clip(params_lower_bound, params_upper_bound, out=pred_params)
    assert pred_params.shape == (len(graphs.rho), 8)
    assert isinstance(graphs.rho[0], np.ndarray)
    assert isinstance(graphs.vp[0], np.ndarray)
    assert isinstance(graphs.rho, List)
    assert isinstance(graphs.vp, List)
    pred_rho = rho_batch(pred_params.tolist(), graphs.rho)
    pred_vp = vp_batch(pred_params.tolist(), graphs.vp)
    assert isinstance(pred_rho[0], np.ndarray)
    assert isinstance(pred_vp[0], np.ndarray)
    assert isinstance(pred_rho, List)
    assert isinstance(pred_vp, List)
    rho = [rho[:, -1] for rho in graphs.rho if rho.shape[0] > 0]
    vp = [vp[:, -1] for vp in graphs.vp if vp.shape[0] > 0]
    mape_den = []
    for pred, exp in zip(pred_rho, rho):
        assert pred.shape == exp.shape
        mape_den += [np.mean(np.abs(pred - exp) / exp).item()]
    mape_den = np.asarray(mape_den).mean().item()
    mape_vp = []
    for pred, exp in zip(pred_vp, vp):
        assert pred.shape == exp.shape
        mape_vp += [np.mean(np.abs(pred - exp) / exp).item()]
    mape_vp = np.asarray(mape_vp).mean().item()
    return mape_den, mape_vp



"""Module for molecular graph design."""

import torch
from rdkit import RDLogger
from torch_geometric.data import Data

from .ogb_utils import smiles2graph  # from ogb.utils.mol import smiles2graph
from .rdkit_util import *  # pylint: disable = unused-wildcard-import, wildcard-import


# pylint: disable = invalid-name
def from_InChI(
    InChI: str,
    with_hydrogen: bool = False,
    kekulize: bool = False,
) -> Data:
    r"""Converts a InChI string to a :class:`torch_geometric.data.Data`
    instance.

    Args:
        InChI (str): The InChI string.
        with_hydrogen (bool, optional): If set to :obj:`True`, will store
            hydrogens in the molecule graph. (default: :obj:`False`)
        kekulize (bool, optional): If set to :obj:`True`, converts aromatic
            bonds to single/double bonds. (default: :obj:`False`)
    """

    RDLogger.DisableLog("rdApp.*")  # type: ignore

    smiles = inchitosmiles(InChI, with_hydrogen, kekulize)
    graph = smiles2graph(smiles)

    x = graph["node_feat"]
    edge_attr = graph["edge_feat"]
    edge_index = graph["edge_index"]

    SELECTED_ATOMS = (6, 7, 8, 9, 14, 15, 16, 17, 35, 53)
    g_atom_count = atom_count(smiles)
    g_atom_count = [g_atom_count[i] for i in SELECTED_ATOMS]

    return Data(
        x=torch.from_numpy(x),
        edge_attr=torch.from_numpy(edge_attr),
        edge_index=torch.from_numpy(edge_index),
        InChI=InChI,
        smiles=smiles,
        ecfp=torch.from_numpy(ECFP(smiles)),
        mw=torch.as_tensor([[mw(InChI)]]),
        ring_count=torch.as_tensor([[ring_count(smiles)]]),
        rbond_count=torch.as_tensor([[rbond_count(smiles)]]),
        atom_count=torch.as_tensor([g_atom_count]),
    )


def from_smiles(smiles: str) -> Data:
    r"""Converts a smile string to a :class:`torch_geometric.data.Data`
    instance.

    Args:
        smile (str): The smile string.
    """

    inchi = smilestoinchi(smiles)
    return from_InChI(inchi)



"""Module to be used for hyperparameter tuning"""

import os
from functools import partial

import torch
from absl import app, flags, logging
from ray import tune
from ray.air.integrations.wandb import WandbLoggerCallback
from ray.tune.schedulers import HyperBandForBOHB
from ray.tune.search import ConcurrencyLimiter
from ray.tune.search.bohb import TuneBOHB

from .search_space import get_search_space
from .train import training_updated
from .utils import CustomStopper

os.environ["WANDB_SILENT"] = "true"
# os.environ["WANDB_MODE"] = "offline"
# os.environ["RAY_AIR_NEW_OUTPUT"] = "0"

FLAGS = flags.FLAGS

flags.DEFINE_string(
    "restoredir",
    None,
    "Directory path to restore the state of a searcher from previous tuning results",
)
flags.DEFINE_string("resumedir", None, "Directory path to resume unfinished tuning")
flags.DEFINE_integer(
    "max_concurrent", 4, "Maximum concurrent samples from the underlying searcher."
)
flags.DEFINE_integer("num_samples", 100, "Number of trials")
flags.DEFINE_float(
    "time_budget_s",
    3600,
    "Global time budget in seconds after which all trials are stopped",
)


def main(argv):
    """Execution from command line"""
    if len(argv) > 1:
        raise app.UsageError("Too many command-line arguments.")

    logging.info("Calling tuner!")
    torch.set_float32_matmul_precision("medium")

    config: dict = FLAGS.config
    # Hyperparameter search space
    search_space = get_search_space()
    max_t = config["num_train_steps"] // config["eval_every_steps"]
    # BOHB search algorithm
    search_alg = TuneBOHB(
        space=search_space,
        metric="mape_den/dataloader_idx_0",
        mode="min",
        points_to_evaluate=[
            {
                "conv": "PNA",
                "global_pool": "add",
                "propagation_depth": 6,
                "hidden_dim": 256,
                "post_layers": 3,
                "pre_layers": 2,
                "towers": 2,
            },
            {
                "conv": "GATv2",
                "global_pool": "add",
                "propagation_depth": 4,
                "hidden_dim": 256,
                "heads": 2,
            },
        ],
        seed=77,
        max_concurrent=FLAGS.max_concurrent,
    )
    search_alg = ConcurrencyLimiter(search_alg, max_concurrent=FLAGS.max_concurrent)
    if FLAGS.restoredir:
        search_alg.restore_from_dir(FLAGS.restoredir)
    # Early stopping scheduler for BOHB
    scheduler = HyperBandForBOHB(
        time_attr="training_iteration",
        metric="mape_den/dataloader_idx_0",
        mode="min",
        max_t=max_t,
        stop_last_trials=True,
    )
    # reporter = TrialTerminationReporter()
    stopper = CustomStopper(max_t)

    trainable = tune.with_resources(
        partial(
            training_updated,
            config=FLAGS.config,
            workdir=FLAGS.workdir,
        ),
        resources={
            "CPU": FLAGS.num_cpu,
            "GPU": FLAGS.num_gpus,
        },
    )

    if FLAGS.resumedir:
        tuner = tune.Tuner.restore(
            FLAGS.resumedir,
            trainable,
            resume_unfinished=True,
            resume_errored=False,
            restart_errored=True,
        )
    else:
        tuner = tune.Tuner(
            trainable,
            tune_config=tune.TuneConfig(
                search_alg=search_alg,
                scheduler=scheduler,
                num_samples=FLAGS.num_samples,
                time_budget_s=FLAGS.time_budget_s,
                reuse_actors=False,
                trial_name_creator=lambda trial: trial.trial_id,
                trial_dirname_creator=lambda trial: trial.trial_id,
            ),
            run_config=tune.RunConfig(
                name="gnnpcsaft",
                storage_path=os.path.join(FLAGS.workdir, "ray_results"),
                checkpoint_config=tune.CheckpointConfig(
                    num_to_keep=1,
                ),
                verbose=0,
                progress_reporter=None,
                log_to_file=False,
                stop=stopper,
                callbacks=(
                    [
                        WandbLoggerCallback(
                            "gnn-pc-saft",
                            config["dataset"],
                            tags=["tuning", config["dataset"]] + FLAGS.tags,
                        )
                    ]
                ),
            ),
        )

    tuner.fit()


if __name__ == "__main__":
    flags.mark_flags_as_required(["config", "workdir"])
    app.run(main)



"Module for training with ECFP"

import os

import torch
import xgboost as xgb
from absl import app, flags, logging
from torch_geometric.loader import DataLoader

from .utils import build_test_dataset, build_train_dataset, rf_xgb_evaluation


def training(workdir: str, config: dict):
    """Training function"""
    # Load the data
    train_dataset = build_train_dataset(workdir, "esper")
    val_dataset, train_val_dataset = build_test_dataset(workdir, train_dataset)
    # Create the Dataloader
    train_loader = DataLoader(
        train_dataset, batch_size=len(train_dataset), shuffle=True
    )
    val_loader = DataLoader(val_dataset, batch_size=len(val_dataset))
    train_val_dataloader = DataLoader(
        train_val_dataset, batch_size=len(train_val_dataset)
    )
    graphs_train = next(iter(train_loader))
    train_dmatrix = xgb.DMatrix(
        torch.hstack(
            (
                graphs_train.ecfp,
                graphs_train.mw,
                graphs_train.atom_count,
                graphs_train.ring_count,
                graphs_train.rbond_count,
            )
        ).numpy(),
        label=graphs_train.para.numpy(),
    )

    xgb_param = {
        "booster": "gblinear",
        "objective": "reg:squaredlogerror",
        "eval_metric": ["mape", "rmsle"],
        "device": "cuda",
        "lambda": config["lambda"],
        "alpha": config["alpha"],
        "eta": config["eta"],
    }
    # Train the model
    results = {}
    xgb_model = xgb.train(
        xgb_param,
        train_dmatrix,
        num_boost_round=config["num_boost_round"],
        evals=[(train_dmatrix, "train")],
        evals_result=results,
        verbose_eval=False,
    )

    print(
        f"Train-mape: {results['train']['mape'][-1]} "
        f"|| Train-rmsle: {results['train']['rmsle'][-1]}"
    )

    # evaluate on validation set
    mape_den, mape_vp = rf_xgb_evaluation(next(iter(val_loader)), xgb_model)

    print(f"MAPE den/val: {mape_den}, MAPE vp/val: {mape_vp}")

    mape_den, mape_vp = rf_xgb_evaluation(next(iter(train_val_dataloader)), xgb_model)

    print(f"MAPE den/train_val: {mape_den}, MAPE vp/train_val: {mape_vp}")

    # save model
    xgb_model.save_model(os.path.join(workdir, "train/checkpoints/xgb_model.json"))
    return mape_den, mape_vp


FLAGS = flags.FLAGS

flags.DEFINE_string("workdir", None, "Working Directory.")
flags.DEFINE_integer("num_boost_round", 1000, "Number of boosting rounds.")


def main(argv):
    """Execution from command line"""
    if len(argv) > 1:
        raise app.UsageError("Too many command-line arguments.")

    logging.info("Calling training!")

    training(
        FLAGS.workdir,
        {
            "lambda": 1.92e-6,
            "alpha": 2.77e-5,
            "num_boost_round": FLAGS.num_boost_round,
            "eta": 0.039,
        },
    )


if __name__ == "__main__":
    flags.mark_flags_as_required(["workdir"])
    app.run(main)



"""Module for calculating molecule complexity and family group."""

from urllib.parse import quote
from urllib.request import HTTPError, urlopen

from rdkit import Chem

# pyright: reportAttributeAccessIssue=false
# pylint: disable = no-name-in-module
from rdkit.Chem.Fragments import (
    fr_Al_OH,
    fr_aldehyde,
    fr_amide,
    fr_Ar_OH,
    fr_benzene,
    fr_COO,
    fr_epoxide,
    fr_ester,
    fr_ether,
    fr_halogen,
    fr_ketone,
    fr_NH0,
    fr_NH1,
    fr_NH2,
    fr_nitrile,
    fr_phenol,
    fr_phos_acid,
    fr_SH,
    fr_sulfide,
    fr_unbrch_alkane,
)


def complexity(ids: str) -> float:
    """Complexity as implemented by `PubChem`."""
    try:
        url = (
            "https://pubchem.ncbi.nlm.nih.gov/"
            + "rest/pug/compound/inchi/property/complexity/TXT?inchi="
            + quote(ids)
        )
        with urlopen(url) as ans:
            ans = ans.read().decode("utf8").rstrip()
            ans = ans.split("\n")[0]
            if ans:
                ans = float(ans)
    except (TypeError, ValueError, HTTPError):
        print("not ok:", ids)
        ans = float("inf")

    # sleep(0.1)
    return ans


def get_family_groups(inchi: str) -> list[str]:
    """Find a family groups for a molecule."""
    mol = Chem.MolFromInchi(inchi, sanitize=True)

    if mol is None:
        mol = Chem.MolFromInchi(inchi, sanitize=False)

    frs_1 = [
        fr_COO,
        fr_ketone,
        fr_ester,
        fr_ether,
        fr_aldehyde,
        fr_NH0,
        fr_NH1,
        fr_NH2,
        fr_amide,
        fr_halogen,
        fr_phenol,
        fr_epoxide,
        fr_sulfide,
        fr_SH,
        fr_nitrile,
        fr_phos_acid,
    ]
    frs_str1 = [
        "carb acid",
        "ketone/aldehyde",
        "ester/ether/epoxide",
        "ester/ether/epoxide",
        "ketone/aldehyde",
        "amine/amide/nitrile",
        "amine/amide/nitrile",
        "amine/amide/nitrile",
        "amine/amide/nitrile",
        "halogen",
        "phenol",
        "ester/ether/epoxide",
        "thiol/thiolether",
        "thiol/thiolether",
        "amine/amide/nitrile",
        "phosphoric acid",
    ]
    frs_2 = [
        fr_Al_OH,
        fr_Ar_OH,
        fr_benzene,
    ]
    frs_str2 = ["hydroxyl", "hydroxyl", "benzene"]

    frags = [fr(mol) for fr in frs_1]
    frs_str_match = []
    for i, frag in enumerate(frags):
        if frag > 0:
            frs_str_match.append(frs_str1[i])

    if len(frs_str_match) == 0:
        frags = [fr(mol) for fr in frs_2]
        frs_str_match = []
        for i, frag in enumerate(frags):
            if frag > 0:
                frs_str_match.append(frs_str2[i])
    if (len(frs_str_match) == 0) & (fr_unbrch_alkane(mol) > 0):
        frs_str_match.append("unbranched alkane")
    if len(frs_str_match) == 0:
        frs_str_match.append("other")

    return list(set(frs_str_match))



"""Module to be used for model training"""

import os
import os.path as osp
from functools import partial
from pathlib import Path
from typing import Any, Union

import lightning as L
import numpy as np
import torch
import wandb
from absl import app, flags, logging
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import WandbLogger
from ml_collections import config_flags
from ray import train, tune
from ray.train.torch import TorchTrainer
from torch_geometric.loader import DataLoader

from ..configs.configs_parallel import get_configs
from .models import GNNePCSAFTL, HabitchNNL, create_model
from .utils import (
    CustomRayTrainReportCallback,
    EpochTimer,
    build_test_dataset,
    build_train_dataset,
    calc_deg,
)


def create_logger(config, dataset):
    "Creates wandb logging or equivalent."
    wandb.init(
        # Set the project where this run will be logged
        project="gnn-pc-saft",
        # Track hyperparameters and run metadata
        config=config.to_dict(),
        group=dataset,
        tags=[dataset, "train"],
        job_type="train",
    )


def ltrain_and_evaluate(  # pylint:  disable=too-many-locals
    config: dict[str, Any], workdir: str
):
    """Execute model training and evaluation loop with lightning.

    Args:
      config: Hyperparameter configuration for training and evaluation.
      workdir: Working Directory.
    """
    torch.set_float32_matmul_precision("medium")
    # Dataset building
    train_dataset = build_train_dataset(workdir, config["dataset"])
    val_dataset, train_val_dataset = build_test_dataset(workdir, train_dataset)

    train_loader = DataLoader(
        train_dataset,
        batch_size=config["batch_size"],
        shuffle=True,
        num_workers=os.cpu_count(),
    )

    val_dataloader = DataLoader(
        val_dataset,
        batch_size=len(val_dataset),
        num_workers=os.cpu_count(),
    )
    train_val_dataloader = DataLoader(
        train_val_dataset,
        batch_size=len(train_val_dataset),
        num_workers=os.cpu_count(),
    )

    # trainer callback and logger
    callbacks, logger = get_callbacks_and_logger(config, workdir)

    # creating model from config
    model = create_model(config, calc_deg(config["dataset"], workdir))
    # model = torch.compile(model, dynamic=True) off for old gpus

    # creating Lighting trainer function
    trainer = L.Trainer(
        devices="auto",
        accelerator=config["accelerator"],
        strategy="auto",
        max_steps=config["num_train_steps"],
        log_every_n_steps=config["log_every_steps"],
        val_check_interval=config["eval_every_steps"],
        check_val_every_n_epoch=None,
        num_sanity_val_steps=0,
        callbacks=callbacks,
        logger=logger,
        plugins=None,
        enable_progress_bar=False,
        enable_checkpointing=config["job_type"] == "train",
    )

    ckpt_path = get_ckpt_path(config, workdir, model, logger)

    # training run
    logging.info("Training run!")
    trainer.fit(
        model,
        train_loader,
        (
            [train_val_dataloader, val_dataloader]
            if config["dataset"] == "esper"
            else [train_val_dataloader, train_val_dataloader]
        ),
        ckpt_path=ckpt_path,
    )

    if config["job_type"] == "train":
        wandb.finish()


def get_ckpt_path(
    config: dict[str, Any],
    workdir: str,
    model: Union[GNNePCSAFTL, HabitchNNL],
    logger: Union[WandbLogger, None],
):
    "gets checkpoint path for resuming training"
    ckpt_path = None
    if config["job_type"] == "tuning":

        checkpoint: tune.Checkpoint = tune.get_checkpoint()
        trial_id = tune.get_context().get_trial_id()

        if checkpoint:
            with checkpoint.as_directory() as ckpt_dir:
                ckpt_path = Path(ckpt_dir) / f"{trial_id}.ckpt"
        elif config["checkpoint"]:
            ckpt_path = Path(workdir) / f"train/checkpoints/{config['checkpoint']}"
    elif config["checkpoint"]:
        if logger:
            ckpt_dir = Path(workdir) / f"train/checkpoints/{config['model_name']}"
            artifact = logger.use_artifact(config["checkpoint"], "model")
            artifact.download(ckpt_dir)
            ckpt_path = ckpt_dir / "model.ckpt"
            if config["change_opt"]:

                ckpt = torch.load(ckpt_path, weights_only=True)
                model.load_state_dict(ckpt["state_dict"])

                ckpt_path = None
    return ckpt_path


def get_callbacks_and_logger(config, workdir):
    """Creates callbacks and logger for training."""
    callbacks = []
    job_type = config.job_type
    dataset = config.dataset

    if job_type == "train":
        # Checkpointing from val loss (mape_den) and train loss (train_mape)
        checkpoint_mape_den = ModelCheckpoint(
            dirpath=osp.join(workdir, "train/checkpoints"),
            filename=config.model_name + "-{epoch}-mape_den_train",
            save_last=False,
            monitor="mape_den/dataloader_idx_0",
            save_top_k=1,
            verbose=True,
        )
        callbacks.append(checkpoint_mape_den)

        checkpoint_mape_den_2 = ModelCheckpoint(
            dirpath=osp.join(workdir, "train/checkpoints"),
            filename=config.model_name + "-{epoch}-mape_den_val",
            save_last=False,
            monitor="mape_den/dataloader_idx_1",
            save_top_k=1,
            verbose=True,
        )
        callbacks.append(checkpoint_mape_den_2)

        checkpoint_train_loss = ModelCheckpoint(
            dirpath=osp.join(workdir, "train/checkpoints"),
            filename=config.model_name + "-{epoch}-{train_mape:.4f}",
            save_last=False,
            monitor="train_mape",
            save_top_k=1,
            verbose=True,
        )
        callbacks.append(checkpoint_train_loss)

        checkpoint_last_epoch = ModelCheckpoint(
            dirpath=osp.join(workdir, "train/checkpoints"),
            filename=config.model_name + "-{epoch}-{step}",
            verbose=True,
        )
        callbacks.append(checkpoint_last_epoch)

        callbacks.append(EpochTimer())

        # Logging training results at wandb
        logger = WandbLogger(
            log_model=True,
            # Set the project where this run will be logged
            project="gnn-pc-saft",
            id=config.resume_id if config.resume_id else None,
            # Track hyperparameters and run metadata
            config=config.to_dict(),
            group=dataset,
            tags=[dataset, "train", config.model_name],
            job_type="train",
            resume="must" if config.resume_id else "allow",
        )
    else:
        callbacks.append(CustomRayTrainReportCallback())
        logger = None
    return callbacks, logger


def training_parallel(
    train_loop_config: dict[str, Any],
    config: dict[str, Any],
    workdir: str,
):
    """Execute model training and evaluation loop in parallel with ray.

    Args:
      train_loop_config: Dict with hyperparameter configuration
      for training and evaluation in each worker.
    """

    local_rank = str(train.get_context().get_local_rank())
    train_config = train_loop_config[local_rank]
    # selected hyperparameters to test
    training_updated(train_config, config, workdir)


def training_updated(
    train_config: dict[str, Any],
    config: dict[str, Any],
    workdir: str,
):
    """Execute model training and evaluation loop with updated config.

    Args:
      train_config: Updated hyperparameter configuration for training and evaluation.

    """

    for hparam in train_config:
        value = train_config[hparam]
        if isinstance(value, (np.signedinteger,)):
            value = int(value)
        if isinstance(value, (np.floating,)):
            value = float(value)
        config[hparam] = value

    ltrain_and_evaluate(config, workdir)


FLAGS = flags.FLAGS

flags.DEFINE_string("workdir", None, "Working Directory.")
flags.DEFINE_string(
    "framework", "lightning", "Define framework to run: lightning or ray."
)
flags.DEFINE_list("tags", [], "wandb tags")
flags.DEFINE_float("num_cpu", 1.0, "Fraction of CPU threads per trial for ray")
flags.DEFINE_float("num_gpus", 1.0, "Fraction of GPUs per trial for ray")
flags.DEFINE_integer("num_workers", 1, "Number of workers for ray")
config_flags.DEFINE_config_file(
    "config",
    None,
    "File path to the training hyperparameter configuration.",
    lock_config=True,
)


def main(argv):
    """Execution from command line"""
    if len(argv) > 1:
        raise app.UsageError("Too many command-line arguments.")

    logging.info("Calling train and evaluate!")
    torch.set_float32_matmul_precision("medium")

    if FLAGS.framework == "lightning":
        ltrain_and_evaluate(FLAGS.config, FLAGS.workdir)
    elif FLAGS.framework == "ray":
        test_configs = get_configs()
        train_loop_config = {
            str(local_rank): test_configs[local_rank]
            for local_rank in range(FLAGS.num_workers)
        }
        scaling_config = train.ScalingConfig(
            num_workers=FLAGS.num_workers,
            use_gpu=True,
            resources_per_worker={"CPU": FLAGS.num_cpu, "GPU": FLAGS.num_gpus},
        )
        run_config = train.RunConfig(
            name="gnnpcsaft",
            storage_path=None,
            progress_reporter=None,
            log_to_file=False,
            stop=None,
            callbacks=(None),
        )
        trainer = TorchTrainer(
            partial(
                training_parallel,
                config=FLAGS.config,
                workdir=FLAGS.workdir,
            ),
            train_loop_config=train_loop_config,
            scaling_config=scaling_config,
            run_config=run_config,
        )

        result = trainer.fit()
        print(result.metrics_dataframe)


if __name__ == "__main__":
    flags.mark_flags_as_required(["config", "workdir"])
    app.run(main)



"Module for helper functions with rdkit"

from typing import List, Tuple

import numpy as np
from rdkit import Chem, RDLogger
from rdkit.Chem import AllChem

# pylint: disable = no-name-in-module
from rdkit.Chem.Fragments import fr_COO2  # type: ignore
from rdkit.Chem.Fragments import fr_Imine, fr_isocyan, fr_isothiocyan  # type: ignore
from rdkit.Chem.rdMolDescriptors import (
    CalcExactMolWt,
    CalcNumHBA,
    CalcNumHBD,
    CalcNumRings,
    CalcNumRotatableBonds,
)

RDLogger.DisableLog("rdApp.*")  # type: ignore


# pylint: disable = invalid-name
def inchitosmiles(
    InChI: str, with_hydrogen: bool = False, kekulize: bool = False
) -> str:
    """Transform InChI to a SMILES.

    Args:
        InChI (str): InChI
        with_hydrogen (bool): Add hydrogens to the molecule
        kekulize (bool): Kekulize the molecule
    """
    mol = Chem.MolFromInchi(InChI)
    if mol is None:
        raise ValueError("InChI is not valid")

    # pylint: disable = no-member
    if with_hydrogen:
        mol = Chem.AddHs(mol)
    if kekulize:
        Chem.Kekulize(mol)

    smiles = Chem.MolToSmiles(mol)
    return smiles


def smilestoinchi(
    smiles: str, with_hydrogen: bool = False, kekulize: bool = False
) -> str:
    """Transform SMILES to InChI.

    Args:
        smiles (str): SMILES
        with_hydrogen (bool): Add hydrogens to the molecule
        kekulize (bool): Kekulize the molecule
    """
    # pylint: disable = no-member
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        raise ValueError("SMILES is not valid")

    if with_hydrogen:
        mol = Chem.AddHs(mol)
    if kekulize:
        Chem.Kekulize(mol)

    inchi = Chem.MolToInchi(mol)
    return inchi  # type: ignore


def assoc_number(inchi: str) -> Tuple[int, int]:
    """Calculates the number of H-bond acceptors/donors

    Args:
        inchi (str): InChI
    """

    exceptions = (
        "InChI=1S/H2O/h1H2",
        "InChI=1S/Cl2/c1-2",
        "InChI=1S/F2/c1-2",
    )  # From Esper et al. (2023) 10.1021/acs.iecr.3c02255
    if inchi in exceptions:
        return 1, 1
    mol = Chem.MolFromInchi(inchi, removeHs=False)
    if mol is None:
        raise ValueError("InChI is not valid")
    mol = Chem.AddHs(mol)
    na = CalcNumHBA(mol)
    nb = CalcNumHBD(mol)
    if na > 0:
        n_coo = fr_COO2(mol)
        n_Imine = fr_Imine(mol)
        n_isocyanates = fr_isocyan(mol)
        n_isothiocyanates = fr_isothiocyan(mol)
        n_priamide = len(
            mol.GetSubstructMatches(
                Chem.AddHs(
                    Chem.MolFromInchi("InChI=1S/CH3NO/c2-1-3/h1H,(H2,2,3)"),
                    onlyOnAtoms=[1, 2],
                )
            )
        )
        n_sulfuro = len(
            mol.GetSubstructMatches(
                Chem.AddHs(Chem.MolFromSmiles("S(=O)(=O)O"), onlyOnAtoms=[3])
            )
        )

        na -= (
            n_coo + n_priamide + n_Imine - n_isocyanates - n_isothiocyanates + n_sulfuro
        )

    return na, nb


def mw(inchi: str) -> float:
    """Calcultes molecular weight.

    Args:
        inchi (str): InChI
    """
    try:
        mol = Chem.MolFromInchi(inchi, removeHs=False, sanitize=False)
        mol_weight = CalcExactMolWt(mol)
    except (TypeError, ValueError):
        mol_weight = 0

    return mol_weight


def ECFP(smiles: str, radius: int = 3, nBits: int = 3072) -> np.ndarray:
    """Calculates ECFP fingerprints.

    Args:
        smiles (str): SMILES
        radius (int): Radius
        nBits (int): Number of bits
    """
    try:
        mol = Chem.MolFromSmiles(smiles, sanitize=True)
        fp = AllChem.GetMorganFingerprintAsBitVect(mol, radius, nBits=nBits)  # type: ignore
        fp = np.array([fp], dtype=np.int8)
    except (TypeError, ValueError):
        fp = np.zeros((1, nBits), dtype=np.int8)
    return fp


def ring_count(smiles: str) -> int:
    """Calculates the number of rings.

    Args:
        smiles (str): SMILES
    """
    try:
        mol = Chem.MolFromSmiles(smiles, sanitize=True)
        _ring_count = CalcNumRings(mol)
    except (TypeError, ValueError):
        _ring_count = 0
    return _ring_count


def rbond_count(smiles: str) -> int:
    """Calculates the number of rotatable bonds.

    Args:
        smiles (str): SMILES
    """
    try:
        mol = Chem.MolFromSmiles(smiles, sanitize=True)
        _rbond_count = CalcNumRotatableBonds(mol)
    except (TypeError, ValueError):
        _rbond_count = 0
    return _rbond_count


def atom_count(smiles: str) -> List[int]:
    """Count the number of each atom in a molecule.

    Args:
        smiles (str): SMILES
    """
    atom_count_list = [0] * 119
    try:
        mol = Chem.MolFromSmiles(smiles, sanitize=True)
        for atom in mol.GetAtoms():
            atom_count_list[atom.GetAtomicNum()] += 1
    except (TypeError, ValueError):
        pass
    return atom_count_list



"""ePC-SAFT implementation with jax
---------------
Code reference: `https://github.com/zmeri/PC-SAFT`
"""

# @author: Wildson Lima


import jax
import jax.numpy as np

from .epcsaft_jax import pcsaft_ares

# pylint: disable=C0103,E1102
dares_drho = jax.jit(jax.jacfwd(pcsaft_ares, 5))


# pylint: disable = invalid-name
@jax.jit
def pcsaft_Z(
    x,
    t,
    rho,
    params,
):
    """
    Calculate the compressibility factor.

    Parameters
    ----------
    x : ndarray, shape (n,1)
        Mole fractions of each component. It has a length of n, where n is
        the number of components in the system.
    m : ndarray, shape (n,1)
        Segment number for each component.
    s : ndarray, shape (n,1)
        Segment diameter for each component. For ions this is the diameter of
        the hydrated ion. Units of Angstrom.
    e : ndarray, shape (n,1)
        Dispersion energy of each component. For ions this is the dispersion
        energy of the hydrated ion. Units of K.
    t : float
        Temperature (K)
    rho : float
        Molar density (mol / m^3)
    k_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for dispersion energy.
        (dimensions: ncomp x ncomp)
    l_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for segment diameter.
        (dimensions: ncomp x ncomp)
    khb_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for association energy.
        (dimensions: ncomp x ncomp)
    e_assoc : ndarray, shape (n,1)
        Association energy of the associating components. For non associating
        compounds this is set to 0. Units of K.
    vol_a : ndarray, shape (n,1)
        Effective association volume of the associating components. For non
        associating compounds this is set to 0.
    dipm : ndarray, shape (n,1)
        Dipole moment of the polar components. For components where the dipole
        term is not used this is set to 0. Units of Debye.
    dip_num : ndarray, shape (n,1)
        The effective number of dipole functional groups on each component
        molecule.
    z : ndarray, shape (n,1)
        Charge number of the ions
    dielc : ndarray, shape (n,1)
        permittivity of each component of the medium to be used for electrolyte
        calculations.

    Returns
    -------
    Z : float
        Compressibility factor
    """

    return (1 + rho * dares_drho(x, t, rho, params)).squeeze()


dares_dx = jax.jit(jax.jacfwd(pcsaft_ares, 0))


@jax.jit
def pcsaft_fugcoef(x, t, rho, params):
    """
    Calculate the fugacity coefficients for one phase of the system.

    Parameters
    ----------
    x : ndarray, shape (n,1)
        Mole fractions of each component. It has a length of n, where n is
        the number of components in the system.
    m : ndarray, shape (n,1)
        Segment number for each component.
    s : ndarray, shape (n,1)
        Segment diameter for each component. For ions this is the diameter of
        the hydrated ion. Units of Angstrom.
    e : ndarray, shape (n,1)
        Dispersion energy of each component. For ions this is the dispersion
        energy of the hydrated ion. Units of K.
    t : float
        Temperature (K)
    rho : float
        Molar density (mol / m^3)
    k_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for dispersion energy.
        (dimensions: ncomp x ncomp)
    l_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for segment diameter.
        (dimensions: ncomp x ncomp)
    khb_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for association energy.
        (dimensions: ncomp x ncomp)
    e_assoc : ndarray, shape (n,1)
        Association energy of the associating components. For non associating
        compounds this is set to 0. Units of K.
    vol_a : ndarray, shape (n,1)
        Effective association volume of the associating components. For non
        associating compounds this is set to 0.
    dipm : ndarray, shape (n,1)
        Dipole moment of the polar components. For components where the dipole
        term is not used this is set to 0. Units of Debye.
    dip_num : ndarray, shape (n,1)
        The effective number of dipole functional groups on each component
        molecule.
    z : ndarray, shape (n,1)
        Charge number of the ions
    dielc : ndarray, shape (n,1)
        permittivity of each component of the medium to be used for electrolyte
        calculations.

    Returns
    -------
    fugcoef : ndarray, shape (n,)
        Fugacity coefficients of each component.
    """

    Z = pcsaft_Z(x, t, rho, params)
    lnZ = np.log(Z)
    ares = pcsaft_ares(x, t, rho, params)

    grad = dares_dx(x, t, rho, params)

    return np.exp(ares + (Z - 1) + grad - (x.T @ grad) - lnZ).squeeze()


@jax.jit
def pcsaft_p(x, t, rho, params):
    """
    Calculate pressure.

    Parameters
    ----------
    x : ndarray, shape (n,1)
        Mole fractions of each component. It has a length of n, where n is
        the number of components in the system.
    m : ndarray, shape (n,1)
        Segment number for each component.
    s : ndarray, shape (n,1)
        Segment diameter for each component. For ions this is the diameter of
        the hydrated ion. Units of Angstrom.
    e : ndarray, shape (n,1)
        Dispersion energy of each component. For ions this is the dispersion
        energy of the hydrated ion. Units of K.
    t : float
        Temperature (K)
    rho : float
        Molar density (mol / m^3)
    k_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for dispersion energy.
        (dimensions: ncomp x ncomp)
    l_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for segment diameter.
        (dimensions: ncomp x ncomp)
    khb_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for association energy.
        (dimensions: ncomp x ncomp)
    e_assoc : ndarray, shape (n,1)
        Association energy of the associating components. For non associating
        compounds this is set to 0. Units of K.
    vol_a : ndarray, shape (n,1)
        Effective association volume of the associating components. For non
        associating compounds this is set to 0.
    dipm : ndarray, shape (n,1)
        Dipole moment of the polar components. For components where the dipole
        term is not used this is set to 0. Units of Debye.
    dip_num : ndarray, shape (n,1)
        The effective number of dipole functional groups on each component
        molecule.
    z : ndarray, shape (n,1)
        Charge number of the ions
    dielc : ndarray, shape (n,1)
        permittivity of each component of the medium to be used for electrolyte
        calculations.

    Returns
    -------
    P : float
        Pressure (Pa)
    """

    kb = 1.380648465952442093e-23  # Boltzmann constant, J K^-1
    nav = 6.022140857e23  # Avogadro's number
    den = rho * nav  # number density, units of Angstrom^-3

    Z = pcsaft_Z(x, t, rho, params)
    P = Z * kb * t * den  # Pa
    return P.squeeze()


@jax.jit
def density_from_nu(nu, t, x, params):
    """
    density calculation from reduced density nu

    """
    m = params["m"]
    s = params["s"]
    e = params["e"]
    nav = 6.022140857e23  # Avogadro's number

    d = s * (1.0 - 0.12 * np.exp(-3.0 * e / t))
    summ = np.sum(x * m * d**3.0)
    return 6.0 / np.pi * nu / summ * 1.0e30 / nav


@jax.jit
def nu_from_density(rho, t, x, params):
    """
    reduced density calculation from density

    """
    m = params["m"]
    s = params["s"]
    e = params["e"]
    nav = 6.022140857e23  # Avogadro's number

    d = s * (1.0 - 0.12 * np.exp(-3.0 * e / t))
    summ = np.sum(x * m * d**3.0)
    return np.pi / 6 * rho * nav / 1.0e30 * summ


@jax.jit
def den_err(nu, x, t, p, params):
    """Find root of this function to calculate the reduced density or pressure."""

    rho_guess = density_from_nu(nu, t, x, params)

    P_fit = pcsaft_p(x, t, rho_guess, params)

    return (P_fit - p) / p


@jax.jit
def den_errSQ(nu, x, t, p, params):
    """Find root of this function to calculate the reduced density or pressure."""

    return den_err(nu, x, t, p, params) ** 2


dden_errSQ_dnu = jax.jit(jax.jacfwd(den_errSQ))


@jax.jit
def pcsaft_den(x, t, p, phase, params):
    """
    Molar density at temperature and pressure given.

    Parameters
    ----------
    x : ndarray, shape (n,1)
        Mole fractions of each component. It has a length of n, where n is
        the number of components in the system.
    m : ndarray, shape (n,1)
        Segment number for each component.
    s : ndarray, shape (n,1)
        Segment diameter for each component. For ions this is the diameter of
        the hydrated ion. Units of Angstrom.
    e : ndarray, shape (n,1)
        Dispersion energy of each component. For ions this is the dispersion
        energy of the hydrated ion. Units of K.
    t : float
        Temperature (K)
    P : float
        Pressure (Pa)
    k_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for dispersion energy.
        (dimensions: ncomp x ncomp)
    l_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for segment diameter.
        (dimensions: ncomp x ncomp)
    khb_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for association energy.
        (dimensions: ncomp x ncomp)
    e_assoc : ndarray, shape (n,1)
        Association energy of the associating components. For non associating
        compounds this is set to 0. Units of K.
    vol_a : ndarray, shape (n,1)
        Effective association volume of the associating components. For non
        associating compounds this is set to 0.
    dipm : ndarray, shape (n,1)
        Dipole moment of the polar components. For components where the dipole
        term is not used this is set to 0. Units of Debye.
    dip_num : ndarray, shape (n,1)
        The effective number of dipole functional groups on each component
        molecule.
    z : ndarray, shape (n,1)
        Charge number of the ions
    dielc : ndarray, shape (n,1)
        permittivity of each component of the medium to be used for electrolyte
        calculations.
    phase : int
        The phase for which the calculation is performed. Options: 1 (liquid),
        0 (vapor).

    Returns
    -------
    rho : float
        Molar density (mol / m^3)
    """

    nulow = (
        10
        ** -np.arange(
            13.0,
            4.0,
            -1,
        )[..., np.newaxis]
    )
    nuhigh = np.arange(
        1.0e-4,
        0.7405,
        0.0001,
    )[..., np.newaxis]
    nu = np.concatenate([nulow, nuhigh], 0)

    err = vden_err(nu, x, t, p, params)

    nul = np.zeros_like(nu).repeat(3, 1) * np.nan

    nul = jax.lax.fori_loop(
        0,
        nul.shape[0] - 1,
        lambda i, nul: jax.lax.cond(
            err[i + 1] * err[i] < 0,
            lambda i, nul: nul.at[i, :].set((nu[i, 0], nu[i + 1, 0], 1)),
            lambda i, nul: nul.at[i, :].set((np.nan, np.nan, 0)),
            i,
            nul,
        ),
        nul,
    )

    nu_max = np.nanargmax(nul, 0)[0]
    nu_min = np.nanargmin(nul, 0)[0]

    nu_max, nu_min = jax.lax.cond(
        phase == 1,
        lambda nul: nul[nu_max, 0:2],
        lambda nul: nul[nu_min, 0:2],
        nul,
    )

    nu = (nu_max + nu_min) / 2.0

    # pylint: disable = unused-argument
    def updater(i, nu):
        f = den_errSQ(nu, x, t, p, params)

        gradf = jax.lax.cond(
            f < 1.0e-5,
            lambda nu: np.inf,
            lambda nu: dden_errSQ_dnu(nu, x, t, p, params),
            nu,
        )

        tmp = nu - f / gradf
        nu = jax.lax.cond(
            np.any((tmp > 0) & (jax.lax.is_finite(tmp))),
            lambda nu: tmp,
            lambda nu: nu,
            nu,
        )

        return nu

    nu = jax.lax.fori_loop(0, 20, updater, nu)

    rho = density_from_nu(nu, t, x, params)

    return rho.squeeze()


vden_err = jax.jit(
    jax.vmap(
        den_err,
        in_axes=(
            0,
            None,
            None,
            None,
        ),
    )
)


def dielc_water(t):
    """
    Return the permittivity of water at 1 bar and the given temperature.

    t : float
        Temperature (K)

    This equation was fit to values given in the reference over the temperature
    range of 263.15 to 368.15 K.

    Reference:
    D. G. Archer and P. Wang, “The permittivity of Water and Debye‐Hückel
    Limiting Law Slopes,” J. Phys. Chem. Ref. Data, vol. 19, no. 2, pp. 371–411,
    Mar. 1990.
    """
    dielc = 7.6555618295e-04 * t**2 - 8.1783881423e-01 * t + 2.5419616803e02
    return dielc


dares_dt = jax.jit(jax.jacfwd(pcsaft_ares, 1))


@jax.jit
def pcsaft_hres(x, t, rho, params):
    """
    Calculate the residual enthalpy for one phase of the system.

    Parameters
    ----------
    x : ndarray, shape (n,1)
        Mole fractions of each component. It has a length of n, where n is
        the number of components in the system.
    m : ndarray, shape (n,1)
        Segment number for each component.
    s : ndarray, shape (n,1)
        Segment diameter for each component. For ions this is the diameter of
        the hydrated ion. Units of Angstrom.
    e : ndarray, shape (n,1)
        Dispersion energy of each component. For ions this is the dispersion
        energy of the hydrated ion. Units of K.
    t : float
        Temperature (K)
    rho : float
        Molar density (mol / m^3)
    k_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for dispersion energy.
        (dimensions: ncomp x ncomp)
    l_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for segment diameter.
        (dimensions: ncomp x ncomp)
    khb_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for association energy.
        (dimensions: ncomp x ncomp)
    e_assoc : ndarray, shape (n,1)
        Association energy of the associating components. For non associating
        compounds this is set to 0. Units of K.
    vol_a : ndarray, shape (n,1)
        Effective association volume of the associating components. For non
        associating compounds this is set to 0.
    dipm : ndarray, shape (n,1)
        Dipole moment of the polar components. For components where the dipole
        term is not used this is set to 0. Units of Debye.
    dip_num : ndarray, shape (n,1)
        The effective number of dipole functional groups on each component
        molecule.
    z : ndarray, shape (n,1)
        Charge number of the ions
    dielc : ndarray, shape (n,1)
        permittivity of each component of the medium to be used for electrolyte
        calculations.

    Returns
    -------
    hres : float
        Residual enthalpy (J mol^-1)
    """

    kb = 1.380648465952442093e-23  # Boltzmann constant, J K^-1
    nav = 6.022140857e23  # Avogadro's number

    grad = dares_dt(x, t, rho, params)

    Z = pcsaft_Z(x, t, rho, params)

    return ((-t * grad + (Z - 1)) * kb * nav * t).squeeze()


@jax.jit
def pcsaft_gres(x, t, rho, params):
    """
    Calculate the residual Gibbs energy for one phase of the system.

    Parameters
    ----------
    x : ndarray, shape (n,1)
        Mole fractions of each component. It has a length of n, where n is
        the number of components in the system.
    m : ndarray, shape (n,1)
        Segment number for each component.
    s : ndarray, shape (n,1)
        Segment diameter for each component. For ions this is the diameter of
        the hydrated ion. Units of Angstrom.
    e : ndarray, shape (n,1)
        Dispersion energy of each component. For ions this is the dispersion
        energy of the hydrated ion. Units of K.
    t : float
        Temperature (K)
    rho : float
        Molar density (mol / m^3)
    k_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for dispersion energy.
        (dimensions: ncomp x ncomp)
    l_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for segment diameter.
        (dimensions: ncomp x ncomp)
    khb_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for association energy.
        (dimensions: ncomp x ncomp)
    e_assoc : ndarray, shape (n,1)
        Association energy of the associating components. For non associating
        compounds this is set to 0. Units of K.
    vol_a : ndarray, shape (n,1)
        Effective association volume of the associating components. For non
        associating compounds this is set to 0.
    dipm : ndarray, shape (n,1)
        Dipole moment of the polar components. For components where the dipole
        term is not used this is set to 0. Units of Debye.
    dip_num : ndarray, shape (n,1)
        The effective number of dipole functional groups on each component
        molecule.
    z : ndarray, shape (n,1)
        Charge number of the ions
    dielc : ndarray, shape (n,1)
        permittivity of each component of the medium to be used for electrolyte
        calculations.

    Returns
    -------
    hres : float
        Residual Gibbs energy (J mol^-1)
    """

    kb = 1.380648465952442093e-23  # Boltzmann constant, J K^-1
    nav = 6.022140857e23  # Avogadro's number

    ares = pcsaft_ares(x, t, rho, params)

    Z = pcsaft_Z(x, t, rho, params)

    return ((ares + (Z - 1) - np.log(Z)) * kb * nav * t).squeeze()


@jax.jit
def pcsaft_sres(x, t, rho, params):
    """
    Calculate the residual entropy (constant volume) for one phase of the system.

    Parameters
    ----------
    x : ndarray, shape (n,1)
        Mole fractions of each component. It has a length of n, where n is
        the number of components in the system.
    m : ndarray, shape (n,1)
        Segment number for each component.
    s : ndarray, shape (n,1)
        Segment diameter for each component. For ions this is the diameter of
        the hydrated ion. Units of Angstrom.
    e : ndarray, shape (n,1)
        Dispersion energy of each component. For ions this is the dispersion
        energy of the hydrated ion. Units of K.
    t : float
        Temperature (K)
    rho : float
        Molar density (mol / m^3)
    k_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for dispersion energy.
        (dimensions: ncomp x ncomp)
    l_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for segment diameter.
        (dimensions: ncomp x ncomp)
    khb_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for association energy.
        (dimensions: ncomp x ncomp)
    e_assoc : ndarray, shape (n,1)
        Association energy of the associating components. For non associating
        compounds this is set to 0. Units of K.
    vol_a : ndarray, shape (n,1)
        Effective association volume of the associating components. For non
        associating compounds this is set to 0.
    dipm : ndarray, shape (n,1)
        Dipole moment of the polar components. For components where the dipole
        term is not used this is set to 0. Units of Debye.
    dip_num : ndarray, shape (n,1)
        The effective number of dipole functional groups on each component
        molecule.
    z : ndarray, shape (n,1)
        Charge number of the ions
    dielc : ndarray, shape (n,1)
        permittivity of each component of the medium to be used for electrolyte
        calculations.

    Returns
    -------
    hres : float
        Residual entropy (J mol^-1)
    """

    hres = pcsaft_hres(x, t, rho, params)

    gres = pcsaft_gres(x, t, rho, params)

    return ((hres - gres) / t).squeeze()


den_phase = jax.jit(
    jax.vmap(
        pcsaft_den,
        in_axes=(
            None,
            None,
            None,
            0,
            None,
        ),
    )
)

fungcoef_phase = jax.jit(
    jax.vmap(
        pcsaft_fugcoef,
        (
            None,
            None,
            0,
            None,
        ),
    )
)


@jax.jit
def k_i(p_guess, x, t, params):
    """Minimize this function to calculate the vapor pressure."""
    phases = np.asarray([1.0, 0.0])

    rho = den_phase(x, t, p_guess, phases, params)

    fugcoef_l, fugcoef_v = fungcoef_phase(
        x,
        t,
        rho,
    )
    return fugcoef_l / fugcoef_v


@jax.jit
def pcsaft_VP(x, t, p_guess, params):
    """
    Vapor pressure calculation

    x : ndarray, shape (n,1)
        Mole fractions of each component. It has a length of n, where n is
        the number of components in the system.
    m : ndarray, shape (n,1)
        Segment number for each component.
    s : ndarray, shape (n,1)
        Segment diameter for each component. For ions this is the diameter of
        the hydrated ion. Units of Angstrom.
    e : ndarray, shape (n,1)
        Dispersion energy of each component. For ions this is the dispersion
        energy of the hydrated ion. Units of K.
    t : float
        Temperature (K)
    k_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for dispersion energy.
        (dimensions: ncomp x ncomp)
    l_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for segment diameter.
        (dimensions: ncomp x ncomp)
    khb_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for association energy.
        (dimensions: ncomp x ncomp)
    e_assoc : ndarray, shape (n,1)
        Association energy of the associating components. For non associating
        compounds this is set to 0. Units of K.
    vol_a : ndarray, shape (n,1)
        Effective association volume of the associating components. For non
        associating compounds this is set to 0.
    dipm : ndarray, shape (n,1)
        Dipole moment of the polar components. For components where the dipole
        term is not used this is set to 0. Units of Debye.
    dip_num : ndarray, shape (n,1)
        The effective number of dipole functional groups on each component
        molecule.
    z : ndarray, shape (n,1)
        Charge number of the ions
    dielc : ndarray, shape (n,1)
        permittivity of each component of the medium to be used for electrolyte
        calculations.

    Returns
    -------
    VP : float
        Vapor Pressure (Pa)
    """

    pref = p_guess - 0.01 * p_guess

    pprime = jax.lax.cond(
        p_guess > 1.0e6,
        lambda p_guess: p_guess - 0.005 * p_guess,
        lambda p_guess: p_guess + 0.01 * p_guess,
        p_guess,
    )

    k = k_i(p_guess, x, t, params)
    kprime = k_i(pprime, x, t, params)

    dlnk_dt = (kprime - k) / (pprime - p_guess)
    wi = x * dlnk_dt / np.sum(x * dlnk_dt)

    kb = np.exp(np.sum(wi * np.log(k)))

    kbprime = np.exp(np.sum(wi * np.log(kprime)))

    B = np.log(kbprime / kb) / (1 / pprime - 1 / p_guess)
    A = np.log(kb) - B * (1 / p_guess - 1 / pref)

    p = 1.0 / (1.0 / pref + (np.log(1.0) - A) / B)

    return p.squeeze()



"""Module for ePC-SAFT calculations."""

from .epcsaft_feos import *  # pylint: disable=wildcard-import,unused-wildcard-import
from .epcsaft_teqp import *  # pylint: disable=wildcard-import,unused-wildcard-import



"""
ePC-SAFT implementation with jax
---------------
Code reference: `https://github.com/zmeri/PC-SAFT`
"""

# @author: Wildson Lima

import jax
import jax.numpy as np

KB = 1.380648465952442093e-23  # Boltzmann constant, J K^-1
N_AV = 6.022140857e23  # Avogadro's number
E_CHRG = 1.6021766208  # elementary charge, units of coulomb / 1e-19
PERM_VAC = 8.854187817  # permittivity in vacuum, C V^-1 Angstrom^-1 / 1e-22
E_CHRG_P10 = 1e-19
PERM_VAC_P10 = 1e-22


# pylint: disable=R0914
@jax.jit
def pcsaft_ares(x, t, rho, params):
    """
    Calculates the residual Helmholtz energy.

    Parameters
    ----------
    x : ndarray, shape (n,1)
        Mole fractions of each component. It has a length of n, where n is
        the number of components in the system.
    m : ndarray, shape (n,1)
        Segment number for each component.
    s : ndarray, shape (n,1)
        Segment diameter for each component. For ions this is the diameter of
        the hydrated ion. Units of Angstrom.
    e : ndarray, shape (n,1)
        Dispersion energy of each component. For ions this is the dispersion
        energy of the hydrated ion. Units of K.
    t : float
        Temperature (K)
    rho : float
        Molar density (mol / m^3)
    k_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for dispersion energy.
        (dimensions: ncomp x ncomp)
    l_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for segment diameter.
        (dimensions: ncomp x ncomp)
    khb_ij : ndarray, shape (n,n)
        Binary interaction parameters between components in the mixture for association energy.
        (dimensions: ncomp x ncomp)
    e_assoc : ndarray, shape (n,1)
        Association energy of the associating components. For non associating
        compounds this is set to 0. Units of K.
    vol_a : ndarray, shape (n,1)
        Effective association volume of the associating components. For non
        associating compounds this is set to 0.
    dipm : ndarray, shape (n,1)
        Dipole moment of the polar components. For components where the dipole
        term is not used this is set to 0. Units of Debye.
    dip_num : ndarray, shape (n,1)
        The effective number of dipole functional groups on each component
        molecule.
    z : ndarray, shape (n,1)
        Charge number of the ions.
    dielc : ndarray, shape (n,1)
        relative permittivity of each component of the medium to be used for electrolyte
        calculations.

    Returns
    -------
    ares : float
        Residual Helmholtz energy (J mol^{-1})
    """
    m = params["m"]
    s = params["s"]
    e = params["e"]
    k_ij = params["k_ij"]
    khb_ij = params["khb_ij"]
    l_ij = params["l_ij"]
    e_assoc = params["e_assoc"]
    vol_a = params["vol_a"]
    dipm = params["dipm"]
    dip_num = params["dip_num"]
    z = params["z"]
    dielc = params["dielc"]
    ncomp = x.shape[0]  # number of components

    d = s * (1.0 - 0.12 * np.exp(-3 * e / t))

    den = rho * N_AV / 1.0e30

    a0 = np.asarray(
        [
            0.910563145,
            0.636128145,
            2.686134789,
            -26.54736249,
            97.75920878,
            -159.5915409,
            91.29777408,
        ]
    )
    a1 = np.asarray(
        [
            -0.308401692,
            0.186053116,
            -2.503004726,
            21.41979363,
            -65.25588533,
            83.31868048,
            -33.74692293,
        ]
    )
    a2 = np.asarray(
        [
            -0.090614835,
            0.452784281,
            0.596270073,
            -1.724182913,
            -4.130211253,
            13.77663187,
            -8.672847037,
        ]
    )
    b0 = np.asarray(
        [
            0.724094694,
            2.238279186,
            -4.002584949,
            -21.00357682,
            26.85564136,
            206.5513384,
            -355.6023561,
        ]
    )
    b1 = np.asarray(
        [
            -0.575549808,
            0.699509552,
            3.892567339,
            -17.21547165,
            192.6722645,
            -161.8264617,
            -165.2076935,
        ]
    )
    b2 = np.asarray(
        [
            0.097688312,
            -0.255757498,
            -9.155856153,
            20.64207597,
            -38.80443005,
            93.62677408,
            -29.66690559,
        ]
    )

    n = np.arange(4)[..., np.newaxis]
    zeta = np.pi / 6.0 * den * (d.T**n) @ (x * m)

    eta = zeta[3]
    m_avg = x.T @ m

    s_ij = (s + s.T) / 2.0 * (1 - l_ij)
    e_ij = np.sqrt(e @ e.T) * (1 - k_ij)
    m2es3 = np.sum((x @ x.T) * (m @ m.T) * (e_ij / t) * s_ij**3)
    m2e2s3 = np.sum((x @ x.T) * (m @ m.T) * (e_ij / t) ** 2 * s_ij**3)
    ghs = (
        1.0 / (1.0 - zeta[3])
        + (d @ d.T) / (d + d.T) * 3.0 * zeta[2] / (1.0 - zeta[3]) ** 2
        + ((d @ d.T) / (d + d.T)) ** 2 * 2.0 * zeta[2] ** 2 / (1.0 - zeta[3]) ** 3
    )

    ares_hs = (
        1
        / zeta[0]
        * (
            3 * zeta[1] * zeta[2] / (1 - zeta[3])
            + zeta[2] ** 3 / (zeta[3] * (1 - zeta[3]) ** 2)
            + (zeta[2] ** 3 / zeta[3] ** 2 - zeta[0]) * np.log(1 - zeta[3])
        )
    )

    a = (
        a0
        + (m_avg - 1.0) / m_avg * a1
        + (m_avg - 1.0) / m_avg * (m_avg - 2.0) / m_avg * a2
    )
    b = (
        b0
        + (m_avg - 1) / m_avg * b1
        + (m_avg - 1.0) / m_avg * (m_avg - 2.0) / m_avg * b2
    )

    idx = np.arange(7)
    i1 = np.sum(a * eta**idx)
    i2 = np.sum(b * eta**idx)
    c1 = 1.0 / (
        1.0
        + m_avg * (8 * eta - 2 * eta**2) / (1 - eta) ** 4
        + (1 - m_avg)
        * (20 * eta - 27 * eta**2 + 12 * eta**3 - 2 * eta**4)
        / ((1 - eta) * (2 - eta)) ** 2
    )

    ghs_diag = np.diagonal(ghs)[..., np.newaxis]

    summ = np.sum((x * (m - 1.0)) * (np.log(ghs_diag)))

    ares_hc = m_avg * ares_hs - summ
    ares_disp = -2 * np.pi * den * i1 * m2es3 - np.pi * den * m_avg * c1 * i2 * m2e2s3

    s_ij_diag = np.diagonal(s_ij)[..., np.newaxis]

    polar_params = (m, s, e, dipm, dip_num, e_ij, s_ij, eta)
    assoc_params = (e_assoc, vol_a, khb_ij, s_ij_diag, s_ij, ghs, ncomp)
    ion_params = (s, z, dielc)

    ares = (
        ares_hc
        + ares_disp
        + ares_polar(x, t, den, polar_params)
        + ares_assoc(x, t, den, assoc_params)
        + ares_ion(x, t, den, ion_params)
    )

    return ares.squeeze()


@jax.jit
def xa_find(xa_guess, delta_ij, den, x):
    """Iterate over this function in order to solve for XA"""

    return 1.0 / (1.0 + den * np.sum(x * xa_guess * delta_ij[..., np.newaxis], axis=1))


def ares_polar(x, t, den, params):
    "Polar term for ePC-SAFT."

    # Dipole term (Gross and Vrabec term) --------------------------------------
    # Gross, Joachim, e Jadran Vrabec. “An Equation-of-State Contribution for
    # Polar Components: Dipolar Molecules”. AIChE Journal 52, nº 3 (2006): 1194–1204.
    #  https://doi.org/10.1002/aic.10683.
    # Held, Christoph, Thomas Reschke, Sultan Mohammad, Armando Luza, e Gabriele Sadowski.
    #  “EPC-SAFT Revised”. Chemical Engineering Research and Design,
    #  Advances in Thermodynamics for Chemical Process and Product Design, 92,
    #  nº 12 (1º de dezembro de 2014): 2884–97. https://doi.org/10.1016/j.cherd.2014.05.017.

    m, s, e, dipm, dip_num, e_ij, s_ij, eta = params

    a0dip = np.asarray([0.3043504, -0.1358588, 1.4493329, 0.3556977, -2.0653308])[
        ..., np.newaxis, np.newaxis
    ]
    a1dip = np.asarray([0.9534641, -1.8396383, 2.0131180, -7.3724958, 8.2374135])[
        ..., np.newaxis, np.newaxis
    ]
    a2dip = np.asarray([-1.1610080, 4.5258607, 0.9751222, -12.281038, 5.9397575])[
        ..., np.newaxis, np.newaxis
    ]
    b0dip = np.asarray([0.2187939, -1.1896431, 1.1626889, 0, 0.0])[
        ..., np.newaxis, np.newaxis
    ]
    b1dip = np.asarray([-0.5873164, 1.2489132, -0.5085280, 0, 0])[
        ..., np.newaxis, np.newaxis
    ]
    b2dip = np.asarray([3.4869576, -14.915974, 15.372022, 0, 0])[
        ..., np.newaxis, np.newaxis
    ]

    c0dip = np.asarray([-0.0646774, 0.1975882, -0.8087562, 0.6902849, 0])[
        ..., np.newaxis, np.newaxis, np.newaxis
    ]
    c1dip = np.asarray([-0.9520876, 2.9924258, -2.3802636, -0.2701261, 0])[
        ..., np.newaxis, np.newaxis, np.newaxis
    ]
    c2dip = np.asarray([-0.6260979, 1.2924686, 1.6542783, -3.4396744, 0])[
        ..., np.newaxis, np.newaxis, np.newaxis
    ]

    idxd = np.arange(5)

    # conversion factor, see the note below Table 2 in Gross and Vrabec 2006
    conv = 7242.702976750923

    dipm_sq = dipm**2 / (m * e * s**3) * conv

    m_ij = np.sqrt(m * m.T)

    m_ij = np.minimum(m_ij, 2)

    adip = (
        a0dip
        + (m_ij - 1) / m_ij * a1dip
        + (m_ij - 1) / m_ij * (m_ij - 2) / m_ij * a2dip
    )
    bdip = (
        b0dip
        + (m_ij - 1) / m_ij * b1dip
        + (m_ij - 1) / m_ij * (m_ij - 2) / m_ij * b2dip
    )

    e_ij_diag = np.diagonal(e_ij)[..., np.newaxis]

    s_ij_diag = np.diagonal(s_ij)[..., np.newaxis]

    etan = (eta**idxd)[..., np.newaxis, np.newaxis]

    j2 = np.sum((adip + bdip * e_ij_diag / t) * etan, axis=0)  # e_ij or e_ij_diag ?

    a2 = np.sum(
        (x * x.T)
        * (e_ij_diag / t * e_ij_diag.T / t)
        * (s_ij_diag**3 * s_ij_diag.T**3)
        / s_ij**3
        * (dip_num * dip_num.T)
        * (dipm_sq * dipm_sq.T)
        * j2
    )

    m_ijk = ((m * m.T)[..., np.newaxis] * m.T) ** (1 / 3.0)

    m_ijk = np.minimum(m_ijk, 2)

    cdip = (
        c0dip
        + (m_ijk - 1) / m_ijk * c1dip
        + (m_ijk - 1) / m_ijk * (m_ijk - 2) / m_ijk * c2dip
    )

    etan = etan[..., np.newaxis]
    j3 = np.sum(cdip * etan, axis=0)

    a3 = np.sum(
        ((x * x.T)[..., np.newaxis] * x.T)
        * ((e_ij_diag / t * e_ij_diag.T / t)[..., np.newaxis] * e_ij_diag.T / t)
        * ((s_ij_diag**3 * s_ij_diag.T**3)[..., np.newaxis] * s_ij_diag.T**3)
        / ((s_ij * s_ij.T)[..., np.newaxis] * s_ij.T)
        * ((dip_num * dip_num.T)[..., np.newaxis] * dip_num.T)
        * ((dipm_sq * dipm_sq.T)[..., np.newaxis] * dipm_sq.T)
        * j3
    )

    a2 = -np.pi * den * a2
    a3 = -4 / 3.0 * np.pi**2 * den**2 * a3

    ares_polar_term = a2 / (1.0 - a3 / a2)

    return jax.lax.cond(
        np.any(jax.lax.is_finite(ares_polar_term)),
        lambda ares_polar_term: ares_polar_term,
        np.zeros_like(ares_polar_term),
        ares_polar_term,
    )


# pylint: disable=invalid-name
def ares_assoc(x, t, den, params):
    "Association term for ePC-SAFT."

    # Association term -------------------------------------------------------
    # 2B association type

    e_assoc, vol_a, khb_ij, s_ij_diag, s_ij, ghs, ncomp = params

    eABij = (e_assoc + e_assoc.T) / 2.0 * (1 - khb_ij)

    volABij = (
        np.sqrt(vol_a @ vol_a.T)
        * (np.sqrt(s_ij_diag @ s_ij_diag.T) / (1 / 2.0 * (s_ij_diag + s_ij_diag.T)))
        ** 3.0
    )

    delta_ij = ghs * volABij * s_ij**3 * (np.exp(eABij / t) - 1.0)

    delta_ij_diag = np.diagonal(delta_ij)[..., np.newaxis] + 1e-30

    XA = (
        np.ones((ncomp, 2))
        * (-1 + np.sqrt(1 + 8 * den * delta_ij_diag))
        / (4 * den * delta_ij_diag)
    )

    XA = jax.lax.fori_loop(
        0, 50, lambda i, XA: (XA + xa_find(XA, delta_ij, den, x)) / 2.0, XA
    )

    ares_assoc_term = np.sum(x * (np.log(XA) - XA / 2.0 + 1 / 2.0))

    return jax.lax.cond(
        np.any(jax.lax.is_finite(ares_assoc_term)),
        lambda ares_assoc_term: ares_assoc_term,
        np.zeros_like(ares_assoc_term),
        ares_assoc_term,
    )


def ares_ion(x, t, den, params):
    "Ion term for ePC-SAFT."

    # Ion term ---------------------------------------------------------------
    s, z, dielc = params
    kb_P10 = 1e-23
    kb = KB / kb_P10
    P10 = E_CHRG_P10**2 / kb_P10 / PERM_VAC_P10
    q = z * E_CHRG
    dielc = x.T @ dielc
    # the inverse Debye screening length. Equation 4 in Held et al. 2008.
    kappa = np.sqrt(den * E_CHRG**2 / kb / t / (dielc * PERM_VAC) * (x.T @ z**2) * P10)
    chi = (
        3.0
        / (kappa * s) ** 3
        * (
            1.5
            + np.log(1 + kappa * s)
            - 2.0 * (1.0 + kappa * s)
            + 0.5 * (1.0 + kappa * s) ** 2
        )
    )
    ares_ion_term = (
        -1
        / 12.0
        / np.pi
        / kb
        / t
        / (dielc * PERM_VAC)
        * np.sum(x * q**2 * chi)
        * kappa
        * P10
    )

    return jax.lax.cond(
        np.any(jax.lax.is_finite(ares_ion_term)),
        lambda ares_ion_term: ares_ion_term,
        np.zeros_like(ares_ion_term),
        ares_ion_term,
    )



"""Module to config traning and model parameters"""

import ml_collections


def get_config():
    """Get the default hyperparameter configuration."""
    config = ml_collections.ConfigDict()
    config.job_type = "train"

    # Optimizer.
    config.optimizer = "adam"
    config.learning_rate = 1e-3
    config.change_opt = False
    config.change_sch = False
    config.weight_decay = 1e-2
    config.momentum = 0.9
    config.patience = 5
    config.warmup_steps = 2

    # Training hyperparameters.

    config.accelerator = "auto"
    config.batch_size = 387 // 4 + 1
    config.num_train_steps = 100_000
    config.log_every_steps = 1000
    config.eval_every_steps = 2500
    config.dataset = "esper_assoc_only"
    config.checkpoint = ""  # wandb artifact path
    config.resume_id = ""  # wandb run id
    config.model = "gnn"
    config.model_name = "gnn_assoc"

    # GNN hyperparameters.
    ## General
    config.conv = "GATv2"
    config.global_pool = "add"
    config.propagation_depth = 3
    config.hidden_dim = 512
    config.dropout = 0.0
    config.add_self_loops = True
    config.num_para = 2
    ## PNA
    config.post_layers = 4
    config.pre_layers = 2
    config.towers = 1  # hidden_dim % towers == 0
    config.deg = []
    ## GatedGraphConv, ARMAConv
    config.num_layers = 2
    config.num_stacks = 2
    ## GAT, GATv2, TransformerConv
    config.heads = 8  # hidden_dim % heads == 0

    return config



"Module to avoid importing ogb for smiles2graph"

# ogb==1.3.6
import numpy as np
from rdkit import Chem

# allowable multiple choice node and edge features
allowable_features = {
    "possible_atomic_num_list": list(range(1, 119)) + ["misc"],
    "possible_chirality_list": [
        "CHI_UNSPECIFIED",
        "CHI_TETRAHEDRAL_CW",
        "CHI_TETRAHEDRAL_CCW",
        "CHI_OTHER",
        "misc",
    ],
    "possible_degree_list": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, "misc"],
    "possible_formal_charge_list": [-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, "misc"],
    "possible_numH_list": [0, 1, 2, 3, 4, 5, 6, 7, 8, "misc"],
    "possible_number_radical_e_list": [0, 1, 2, 3, 4, "misc"],
    "possible_hybridization_list": ["SP", "SP2", "SP3", "SP3D", "SP3D2", "misc"],
    "possible_is_aromatic_list": [False, True],
    "possible_is_in_ring_list": [False, True],
    "possible_bond_type_list": ["SINGLE", "DOUBLE", "TRIPLE", "AROMATIC", "misc"],
    "possible_bond_stereo_list": [
        "STEREONONE",
        "STEREOZ",
        "STEREOE",
        "STEREOCIS",
        "STEREOTRANS",
        "STEREOANY",
    ],
    "possible_is_conjugated_list": [False, True],
}


def atom_to_feature_vector(atom):
    """
    Converts rdkit atom object to feature list of indices
    :param mol: rdkit atom object
    :return: list
    """
    atom_feature = [
        safe_index(allowable_features["possible_atomic_num_list"], atom.GetAtomicNum()),
        safe_index(
            allowable_features["possible_chirality_list"], str(atom.GetChiralTag())
        ),
        safe_index(allowable_features["possible_degree_list"], atom.GetTotalDegree()),
        safe_index(
            allowable_features["possible_formal_charge_list"], atom.GetFormalCharge()
        ),
        safe_index(allowable_features["possible_numH_list"], atom.GetTotalNumHs()),
        safe_index(
            allowable_features["possible_number_radical_e_list"],
            atom.GetNumRadicalElectrons(),
        ),
        safe_index(
            allowable_features["possible_hybridization_list"],
            str(atom.GetHybridization()),
        ),
        allowable_features["possible_is_aromatic_list"].index(atom.GetIsAromatic()),
        allowable_features["possible_is_in_ring_list"].index(atom.IsInRing()),
    ]
    return atom_feature


def bond_to_feature_vector(bond):
    """
    Converts rdkit bond object to feature list of indices
    :param mol: rdkit bond object
    :return: list
    """
    bond_feature = [
        safe_index(
            allowable_features["possible_bond_type_list"], str(bond.GetBondType())
        ),
        allowable_features["possible_bond_stereo_list"].index(str(bond.GetStereo())),
        allowable_features["possible_is_conjugated_list"].index(bond.GetIsConjugated()),
    ]
    return bond_feature


def safe_index(l, e):
    """
    Return index of element e in list l. If e is not present, return the last index
    """
    try:
        return l.index(e)
    except (IndexError, ValueError):
        return len(l) - 1


def smiles2graph(smiles_string):
    """
    Converts SMILES string to graph Data object

    Args:
      SMILES_string (str): SMILES

    Returns:
      graph object (dict)
    """

    mol = Chem.MolFromSmiles(smiles_string)
    if mol is None:
        raise ValueError("SMILES is not valid")

    # atoms
    atom_features_list = []
    for atom in mol.GetAtoms():
        atom_features_list.append(atom_to_feature_vector(atom))
    x = np.array(atom_features_list, dtype=np.int64)

    # bonds
    num_bond_features = 3  # bond type, bond stereo, is_conjugated
    if len(mol.GetBonds()) > 0:  # mol has bonds
        edges_list = []
        edge_features_list = []
        for bond in mol.GetBonds():
            i = bond.GetBeginAtomIdx()
            j = bond.GetEndAtomIdx()

            edge_feature = bond_to_feature_vector(bond)

            # add edges in both directions
            edges_list.append((i, j))
            edge_features_list.append(edge_feature)
            edges_list.append((j, i))
            edge_features_list.append(edge_feature)

        # data.edge_index: Graph connectivity in COO format with shape [2, num_edges]
        edge_index = np.array(edges_list, dtype=np.int64).T

        # data.edge_attr: Edge feature matrix with shape [num_edges, num_edge_features]
        edge_attr = np.array(edge_features_list, dtype=np.int64)

    else:  # mol has no bonds
        edge_index = np.empty((2, 0), dtype=np.int64)
        edge_attr = np.empty((0, num_bond_features), dtype=np.int64)

    graph = {}
    graph["edge_index"] = edge_index
    graph["edge_feat"] = edge_attr
    graph["node_feat"] = x
    graph["num_nodes"] = len(x)

    return graph



"""Module to tune xgb"""

import ConfigSpace as CS
from absl import app, flags, logging
from ray import tune
from ray.air.integrations.wandb import WandbLoggerCallback
from ray.tune.schedulers import HyperBandForBOHB
from ray.tune.search import ConcurrencyLimiter
from ray.tune.search.bohb import TuneBOHB

from .xgb_training import training

FLAGS = flags.FLAGS
flags.DEFINE_float("num_cpu", 1.0, "Fraction of CPU threads per trial for ray")
flags.DEFINE_float("num_gpus", 1.0, "Fraction of GPUs per trial for ray")
flags.DEFINE_integer(
    "max_concurrent", 2, "Maximum concurrent samples from the underlying searcher."
)


def main(argv):
    """Execution from command line"""
    if len(argv) > 1:
        raise app.UsageError("Too many command-line arguments.")

    logging.info("Calling tuner!")
    # Hyperparameter search space
    search_space = CS.ConfigurationSpace()
    search_space.add(
        CS.UniformFloatHyperparameter("eta", lower=1e-4, upper=1e-1, log=True),
        CS.UniformFloatHyperparameter("lambda", lower=1e-6, upper=1e-1, log=True),
        CS.UniformFloatHyperparameter("alpha", lower=1e-6, upper=1e-1, log=True),
        CS.Constant("num_boost_round", 5000),
    )

    workdir = FLAGS.workdir
    # BOHB search algorithm
    search_alg = TuneBOHB(
        space=search_space,
        metric="mape_den",
        mode="min",
        seed=77,
        max_concurrent=FLAGS.max_concurrent,
    )
    search_alg = ConcurrencyLimiter(search_alg, max_concurrent=FLAGS.max_concurrent)
    # Early stopping scheduler for BOHB
    scheduler = HyperBandForBOHB(
        metric="mape_den",
        mode="min",
        max_t=1,
        stop_last_trials=True,
    )

    # tune run config with wandb logger
    run_config = tune.RunConfig(
        name="gnnpcsaft",
        storage_path=None,
        callbacks=(
            [
                WandbLoggerCallback(
                    "gnn-pc-saft",
                    "xgb",
                    tags=["tuning", "xgb", "esper"],
                )
            ]
        ),
    )

    def tune_xgb(config: dict):
        """trainable fun for the xgb model"""
        mape_den, mape_vp = training(workdir, config)

        tune.report({"mape_den": mape_den, "mape_vp": mape_vp, "done": True})

    # Run the tuner
    tuner = tune.Tuner(
        tune.with_resources(
            tune_xgb, resources={"cpu": FLAGS.num_cpu, "gpu": FLAGS.num_gpus}
        ),
        tune_config=tune.TuneConfig(
            num_samples=60, search_alg=search_alg, scheduler=scheduler
        ),
        run_config=run_config,
    )
    result = tuner.fit()
    best_trial = result.get_best_result(
        metric="mape_den",
        mode="min",
    )

    print(f"\nBest trial config:\n {best_trial.config}")
    print(f"\nBest trial final metrics:\n {best_trial.metrics}")


if __name__ == "__main__":
    flags.mark_flags_as_required(["workdir"])
    app.run(main)



"Module for training with ECFP"

import os

import joblib
import numpy as np
import torch
from absl import app, flags, logging
from sklearn.ensemble import RandomForestRegressor
from torch_geometric.loader import DataLoader

from .utils import build_test_dataset, build_train_dataset, rf_xgb_evaluation


def training(workdir: str, config: dict):
    """Training function"""
    # Load the data
    train_dataset = build_train_dataset(workdir, "esper")
    test_dts = build_test_dataset(workdir, train_dataset)
    # Create the Dataloader
    train_loader = DataLoader(
        train_dataset, batch_size=len(train_dataset), shuffle=True
    )
    val_loader = DataLoader(test_dts[0], batch_size=len(test_dts[0]))
    train_val_dataloader = DataLoader(test_dts[1], batch_size=len(test_dts[1]))

    rf_model = RandomForestRegressor(
        criterion="squared_error",
        n_estimators=config["n_estimators"],
        max_depth=config["max_depth"],
        min_samples_split=2,
        min_samples_leaf=1,
        max_features=config["max_features"],
        bootstrap=True,
        oob_score=True,
    )

    graphs_train = next(iter(train_loader))
    x = torch.hstack(
        (
            graphs_train.ecfp,
            graphs_train.mw,
            graphs_train.atom_count,
            graphs_train.ring_count,
            graphs_train.rbond_count,
        )
    ).numpy()
    y = graphs_train.para.numpy()

    rf_model.fit(x, y)

    para_pred = rf_model.predict(x)

    para_pred_mape = np.mean(np.abs(para_pred - y) / y)

    print(f"Train-mape: {para_pred_mape.item()}")
    # evaluate on validation set
    mape_den, mape_vp = rf_xgb_evaluation(next(iter(val_loader)), rf_model)

    print(f"MAPE den/val: {mape_den}, MAPE vp/val: {mape_vp}")

    mape_den, mape_vp = rf_xgb_evaluation(next(iter(train_val_dataloader)), rf_model)

    print(f"MAPE den/train_val: {mape_den}, MAPE vp/train_val: {mape_vp}")

    # save model
    joblib.dump(
        rf_model,
        os.path.join(workdir, "train/checkpoints/rf_model.joblib"),
        compress=3,
    )
    return mape_den, mape_vp


FLAGS = flags.FLAGS

flags.DEFINE_string("workdir", None, "Working Directory.")


def main(argv):
    """Execution from command line"""
    if len(argv) > 1:
        raise app.UsageError("Too many command-line arguments.")

    logging.info("Calling training!")

    training(
        FLAGS.workdir,
        {
            "n_estimators": 1000,
            "max_depth": None,
            "max_features": "log2",
        },
    )


if __name__ == "__main__":
    flags.mark_flags_as_required(["workdir"])
    app.run(main)



"Module to calculate properties with ePC-SAFT using FEOS."

from typing import Dict, List, Optional, Tuple

import numpy as np
import si_units as si
from feos import dft  # type: ignore # pylint: disable = E0401
from feos.eos import (  # type: ignore # pylint: disable = E0401
    Contributions,
    EquationOfState,
    PhaseDiagram,
    PhaseEquilibrium,
    State,
)
from feos.pcsaft import (  # type: ignore # pylint: disable = E0401
    Identifier,
    PcSaftParameters,
    PcSaftRecord,
    PureRecord,
)


def pc_saft(parameters: List[float]) -> EquationOfState.pcsaft:
    """Returns a ePC-SAFT equation of state.

    Args:
        parameters: A list with
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb]`

    """
    parameters.append(1.0)
    return pc_saft_mixture([parameters])


def pc_saft_mixture(
    mixture_parameters: List[List[float]],
    kij_matrix: Optional[List[List[float]]] = None,
) -> EquationOfState.pcsaft:
    """Returns a ePC-SAFT equation of state.

    Args:
        mixture_parameters: A list of
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb, MW]`
         for each component of the mixture
        kij_matrix: A matrix of binary interaction parameters
    """
    records = get_records(mixture_parameters)

    if kij_matrix:
        binary_records = np.asarray(kij_matrix)
    else:
        binary_records = np.zeros((len(records), len(records)))
    pcsaftparameters = PcSaftParameters.from_records(
        records, binary_records=binary_records
    )
    eos = EquationOfState.pcsaft(pcsaftparameters)
    return eos


def get_records(mixture_parameters: List[List[float]]) -> list[PureRecord]:
    """Returns a list of `feos.pcsaft.PureRecord`.

    Args:
        mixture_parameters: A list of
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb, MW]`
         for each component of the mixture
    """
    records = []
    for idx, mol_parameters in enumerate(mixture_parameters):
        records.append(
            PureRecord(
                identifier=Identifier(
                    smiles=f"SMILES_{idx}",
                    inchi=f"InChI_{idx}",
                ),
                molarweight=mol_parameters[-1],  # g/mol
                model_record=PcSaftRecord(
                    m=mol_parameters[0],  # units
                    sigma=mol_parameters[1],  # Å
                    epsilon_k=mol_parameters[2],  # K
                    kappa_ab=mol_parameters[3],
                    epsilon_k_ab=mol_parameters[4],  # K
                    mu=mol_parameters[5],  # Debye
                    na=mol_parameters[6],
                    nb=mol_parameters[7],
                ),
            )
        )

    return records


def mix_den_feos(
    parameters: List[List[float]],
    state: List[float],
    kij_matrix: Optional[List[List[float]]] = None,
) -> float:
    """Calculates mixture liquid density (mol/m³) with ePC-SAFT.

    Args:
        parameters: A list of
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb, MW]`
         for each component of the mixture
        state: A list with
         `[Temperature (K), Pressure (Pa), mole_fractions_1, mole_fractions_2, ...]`
        kij_matrix: A matrix of binary interaction parameters
    """

    t = state[0]  # Temperature, K
    p = state[1]  # Pa
    x = np.asarray(state[2:])  # mole fractions

    eos = pc_saft_mixture(parameters, kij_matrix)

    statenpt = State(
        eos,
        temperature=t * si.KELVIN,
        pressure=p * si.PASCAL,
        molefracs=x,
        density_initialization="liquid",
    )

    den = statenpt.density * (si.METER**3) / si.MOL

    return den


def pure_den_feos(parameters: List[float], state: List[float]) -> float:
    """Calculates pure component liquid density (mol/m³) with ePC-SAFT.

    Args:
        parameters: A list with
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb]`
        state: A list with `[Temperature (K), Pressure (Pa)]`
    """

    t = state[0]  # Temperature, K
    p = state[1]  # Pa

    eos = pc_saft(parameters)
    statenpt = State(
        eos,
        temperature=t * si.KELVIN,
        pressure=p * si.PASCAL,
        density_initialization="liquid",
    )

    den = statenpt.density * (si.METER**3) / si.MOL

    return den


def mix_vp_feos(
    parameters: List[List[float]],
    state: List[float],
    kij_matrix: Optional[List[List[float]]] = None,
) -> Tuple[float, float]:
    """Calculates mixture `(Bubble point (Pa), Dew point (Pa))` with ePC-SAFT.

    Args:
        parameters: A list of
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb, MW]`
         for each component of the mixture
        state: A list with
         `[Temperature (K), Pressure (Pa), mole_fractions_1, molefractions_2, ...]`
        kij_matrix: A matrix of binary interaction parameters
    """

    t = state[0]  # Temperature, K
    x = np.asarray(state[2:])  # mole fractions

    eos = pc_saft_mixture(parameters, kij_matrix=kij_matrix)

    vle_bubble_point = PhaseEquilibrium.bubble_point(
        eos, temperature_or_pressure=t * si.KELVIN, liquid_molefracs=x
    )

    vle_dew_point = PhaseEquilibrium.dew_point(
        eos, temperature_or_pressure=t * si.KELVIN, vapor_molefracs=x
    )

    assert (
        t == vle_bubble_point.liquid.temperature / si.KELVIN
    ), "Temperature mismatch for bubble point"
    assert (
        t == vle_dew_point.vapor.temperature / si.KELVIN
    ), "Temperature mismatch for dew point"
    return (
        vle_bubble_point.liquid.pressure() / si.PASCAL,
        vle_dew_point.vapor.pressure() / si.PASCAL,
    )


def pure_vp_feos(parameters: List[float], state: List[float]) -> float:
    """Calculates pure component vapor pressure (Pa) with ePC-SAFT.

    Args:
        parameters: A list with
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb]`
        state: A list with `[Temperature (K)]`
    """

    t = state[0]  # Temperature, K

    eos = pc_saft(parameters)
    vle = PhaseEquilibrium.pure(eos, temperature_or_pressure=t * si.KELVIN)

    assert t == vle.liquid.temperature / si.KELVIN

    return vle.liquid.pressure() / si.PASCAL


def pure_h_lv_feos(parameters: List[float], state: List[float]) -> float:
    """Calculates pure component enthalpy of vaporization (kJ/mol) with ePC-SAFT.

    Args:
        parameters: A list with
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb]`
        state: A list with `[Temperature (K)]`
    """

    t = state[0]  # Temperature, K

    eos = pc_saft(parameters)
    vle = PhaseEquilibrium.pure(eos, temperature_or_pressure=t * si.KELVIN)

    liquid_state = vle.liquid
    vapor_state = vle.vapor

    assert t == liquid_state.temperature / si.KELVIN

    return (
        vapor_state.molar_enthalpy(Contributions.Residual)
        - liquid_state.molar_enthalpy(Contributions.Residual)
    ) * (si.MOL / si.KILO / si.JOULE)


def pure_s_lv_feos(parameters: List[float], state: List[float]) -> float:
    """Calcules pure component entropy of vaporization (J/mol*K) with ePC-SAFT.

    Args:
        parameters: A list with
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb]`
        state: A list with `[Temperature (K)]`
    """
    t = state[0]  # Temperature, K
    eos = pc_saft(parameters)
    vle = PhaseEquilibrium.pure(eos, temperature_or_pressure=t * si.KELVIN)
    liquid_state = vle.liquid
    vapor_state = vle.vapor
    assert t == liquid_state.temperature / si.KELVIN
    return (
        vapor_state.molar_entropy(Contributions.Residual)
        - liquid_state.molar_entropy(Contributions.Residual)
    ) * (si.MOL * si.KELVIN / si.JOULE)


def critical_points_feos(parameters: List[float]) -> List[float]:
    """Calculates critical points `[Tc (K), Pc (Pa), Dc (mol/m³)]` with ePC-SAFT.

    Args:
        parameters: A list with
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb]`
    """
    eos = pc_saft(parameters)
    critical_point = State.critical_point(eos)
    return [
        critical_point.temperature / si.KELVIN,
        critical_point.pressure() / si.PASCAL,
        critical_point.density * (si.METER**3) / si.MOL,
    ]


def pure_viscosity_feos(parameters: List[float], state: List[float]) -> float:
    """Calcules pure component viscosity (kPa*s) with ePC-SAFT.

    Args:
        parameters: A list with
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb]`
        state: A list with `[Temperature (K), Pressure (Pa)]`
    """
    t = state[0]  # Temperature, K
    p = state[1]  # Pa

    eos = pc_saft(parameters)
    statenpt = State(
        eos,
        temperature=t * si.KELVIN,
        pressure=p * si.PASCAL,
        density_initialization="liquid",
    )

    return statenpt.viscosity()  # / (KILO * PASCAL * SECOND)


def phase_diagram_feos(parameters: List[float], state: List[float]) -> Dict[str, float]:
    """Calculates phase diagram from
    state temperature up to the critical temperature with ePC-SAFT.

    Args:
        parameters: A list with
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb]`
        state: A list with `[Temperature (K)]`
    """
    t = state[0]  # Temperature, K
    eos = pc_saft(parameters)
    phase_diagram = PhaseDiagram.pure(eos, min_temperature=t * si.KELVIN, npoints=200)

    return phase_diagram.to_dict(Contributions.Residual)


def pure_surface_tension_feos(
    parameters: List[float], state: List[float]
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculates pure component `[Surface Tension (mN/m), Temperature (K)]` with ePC-SAFT
    from state temperature up to the critical temperature with ePC-SAFT.

    Args:
        parameters: A list with
         `[m, sigma, epsilon/kB, kappa_ab, epsilon_ab/kB, dipole moment, na, nb, MW]`
        state: A list with `[Temperature (K)]`
    """
    t = state[0]  # Temperature, K
    records = get_records([parameters])

    pcsaftparameters = PcSaftParameters.from_records(records)
    functional = dft.HelmholtzEnergyFunctional.pcsaft(pcsaftparameters)
    phase_diagram = dft.PhaseDiagram.pure(functional, t * si.KELVIN, 100)
    st_diagram = dft.SurfaceTensionDiagram(phase_diagram.states, n_grid=1024)

    st = st_diagram.surface_tension / (si.MILLI * si.NEWTON / si.METER)
    temp = st_diagram.liquid.temperature / si.KELVIN
    return st, temp


def parameters_gc_pcsaft(smiles: str) -> List[float]:
    """Calculates PC-SAFT parameters with Group Contribution method.

    Args:
        smiles (str): SMILES of the compound
    """
    pure_record = (
        PcSaftParameters.from_json_smiles(
            [smiles],
            "./gnnepcsaft/data/gc_pcsaft/sauer2014_smarts.json",
            "./gnnepcsaft/data/gc_pcsaft/rehner2023_hetero.json",
        )
        .pure_records[0]
        .model_record
    )

    m = pure_record.m
    sigma = pure_record.sigma
    e = pure_record.epsilon_k
    mu = pure_record.mu if pure_record.mu else 0
    kab = pure_record.kappa_ab if pure_record.kappa_ab else 0
    eab = pure_record.epsilon_k_ab if pure_record.epsilon_k_ab else 0
    na = pure_record.na if pure_record.na else 0
    nb = pure_record.nb if pure_record.nb else 0

    return [m, sigma, e, kab, eab, mu, na, nb]
